"""App related tests."""

from unittest import mock

from django.test import SimpleTestCase
from django.urls import reverse
from django.utils import timezone

from dns.rdtypes.ANY.TXT import TXT

from modoboa.admin import factories as admin_factories
from modoboa.admin import models as admin_models
from modoboa.lib.tests import ModoTestCase

from . import factories
from . import lib


BAD_SPF_RECORDS = [
    ("a mx ~all", "Not an SPF record"),
    ("v=spf1 toto ~all", "Unknown mechanism toto"),
    ("v=spf1 ip4 -all", "Wrong ip4 mechanism syntax"),
    ("v=spf1 +ip4:1.1.1. +all", "Wrong IPv4 address format"),
    ("v=spf1 ip6 ~all", "Wrong ip6 mechanism syntax"),
    ("v=spf1 +ip6:x:: +all", "Wrong IPv6 address format"),
    ("v=spf1 a|domain.com ~all", "Invalid syntax for a mechanism"),
    ("v=spf1 a:domain.com/TOTO", "Invalid mask found TOTO"),
    ("v=spf1 a/1000", "Invalid mask found 1000"),
    ("v=spf1 mx:domain.com/TOTO", "Invalid mask found TOTO"),
    ("v=spf1 mx/1000", "Invalid mask found 1000"),
    ("v=spf1 ptr|test.com", "Invalid syntax for ptr mechanism"),
    ("v=spf1 forward=test.com ~all", "Unknown modifier forward"),
    ("v=spf1 redirect=test.com redirect=test.com ~all",
     "Duplicate modifier redirect found"),
    ("v=spf1", "No mechanism found"),
]

GOOD_SPF_RECORDS = [
    "v=spf1 mx",
    "v=spf1 a ~all",
    "v=spf1 ip4:1.2.3.4 -all",
    "v=spf1 ip4:1.2.3.4/16 -all",
    "v=spf1 ip6:fe80::9eb6:d0ff:fe8e:2807 -all",
    "v=spf1 ip6:fe80::9eb6:d0ff:fe8e:2807/64 -all",
    "v=spf1 a -all",
    "v=spf1 a:example.com -all",
    "v=spf1 a:mailers.example.com -all",
    "v=spf1 a/24 a:offsite.example.com/24 -all",
    "v=spf1 mx mx:deferrals.domain.com -all",
    "v=spf1 mx/24 mx:offsite.domain.com/24 -all",
    "v=spf1 ptr -all",
    "v=spf1 ptr:otherdomain.com -all",
    "v=spf1 exists:example.com -all",
    "v=spf1 include:example.com  -all",
    "v=spf1 ?include:example.com -all",
    "v=spf1 redirect=example.com",
    "v=spf1 exp=example.com",
]

BAD_DMARC_RECORDS = [
    ("", "Not a valid DMARC record"),
    ("v=DMARC1; test", "Invalid tag test"),
    ("v=DMARC1; test=toto", "Unknown tag test"),
    ("v=DMARC1; adkim=g", "Wrong value g for tag adkim"),
    ("v=DMARC1; rua=mail:toto@test.com",
     "Wrong value mail:toto@test.com for tag rua"),
    ("v=DMARC1; rf=afrf,toto", "Wrong value toto for tag rf"),
    ("v=DMARC1; ri=XXX", "Wrong value XXX for tag ri: not an integer"),
    ("v=DMARC1; pct=-1", "Wrong value -1 for tag pct: less than 0"),
    ("v=DMARC1; pct=1000", "Wrong value 1000 for tag pct: greater than 100"),
    ("v=DMARC1; pct=100", "Missing required p tag"),
]

GOOD_DMARC_RECORDS = [
    "v=DMARC1; p=reject;; pct=100",
    "v=DMARC1; p=quarantine; sp=none; adkim=s; aspf=s; "
    "rua=mailto:dmarc-aggrep@ngyn.org; ruf=mailto:toto@test.com!24m; "
    "rf=afrf; pct=100; ri=86400"
]

BAD_DKIM_RECORDS = [
    ("", "Not a valid DKIM record"),
    ("v=DKIM1; toto;p=XXX", "Invalid tag toto"),
    ("v=DKIM1;;k=rsa", "No key found in record"),
]


class LibTestCase(SimpleTestCase):
    """TestCase for library methods."""

    @mock.patch('modoboa.admin.lib.get_dns_records')
    def test_get_record_type_value(self, mock_get_dns_records):
        mock_get_dns_records.return_value = [
            TXT("IN", "TXT", ["v=spf1 mx -all"]),
            TXT("IN", "TXT", ["v=DKIM1 p=XXXXX", "YYYYY"]),
            TXT("IN", "TXT", ["v=DMARC1 p=reject"]),
        ]
        self.assertEqual(
            lib.get_spf_record("example.com"), "v=spf1 mx -all"
        )
        self.assertEqual(
            lib.get_dkim_record("example.com", "mail"), "v=DKIM1 p=XXXXXYYYYY"
        )
        self.assertEqual(
            lib.get_dmarc_record("example.com"), "v=DMARC1 p=reject"
        )

    def test_check_spf_syntax(self):
        for record in BAD_SPF_RECORDS:
            with self.assertRaises(lib.DNSSyntaxError) as ctx:
                lib.check_spf_syntax(record[0])
            self.assertEqual(str(ctx.exception), record[1])
        for record in GOOD_SPF_RECORDS:
            lib.check_spf_syntax(record)

    def test_check_dmarc_syntax(self):
        for record in BAD_DMARC_RECORDS:
            with self.assertRaises(lib.DNSSyntaxError) as ctx:
                lib.check_dmarc_syntax(record[0])
            self.assertEqual(str(ctx.exception), record[1])
        for record in GOOD_DMARC_RECORDS:
            lib.check_dmarc_syntax(record)

    def test_check_dkim_syntax(self):
        for record in BAD_DKIM_RECORDS:
            with self.assertRaises(lib.DNSSyntaxError) as ctx:
                lib.check_dkim_syntax(record[0])
            self.assertEqual(str(ctx.exception), record[1])
        key = lib.check_dkim_syntax("v=DKIM1;p=XXX123")
        self.assertEqual(key, "XXX123")


class ViewsTestCase(ModoTestCase):
    """A test case for views."""

    @classmethod
    def setUpTestData(cls):
        """Create some records."""
        super(ViewsTestCase, cls).setUpTestData()
        cls.spf_rec = factories.DNSRecordFactory(
            type="spf", value="v=SPF1 mx -all", is_valid=True,
            domain__name="test.com"
        )
        cls.dmarc_rec = factories.DNSRecordFactory(
            type="dmarc", value="XXX", is_valid=False,
            error="Not a DMARC record",
            domain__name="test.com"
        )
        cls.dkim_rec = factories.DNSRecordFactory(
            type="dkim", value="12345", is_valid=False,
            error="Public key mismatchs",
            domain__name="test.com"
        )
        cls.ac_rec = factories.DNSRecordFactory(
            type="autoconfig", value="1.2.3.4", is_valid=True,
            domain__name="test.com"
        )

    def test_record_detail_view(self):
        url = reverse("dnstools:dns_record_detail", args=[self.spf_rec.pk])
        response = self.client.get(url)
        self.assertContains(
            response, "A DNS record has been found and is valid")
        url = reverse("dnstools:dns_record_detail", args=[self.dmarc_rec.pk])
        response = self.client.get(url)
        self.assertContains(response, "Not a DMARC record")

    def test_autoconfig_record_status_view(self):
        url = reverse(
            "dnstools:autoconfig_records_status", args=[self.ac_rec.domain.pk])
        response = self.client.get(url)
        self.assertContains(response, "autoconfig record (Mozilla) found")
        self.assertContains(
            response, "autodiscover record (Microsoft) not found")

    def test_domain_dns_configuration(self):
        url = reverse(
            "dnstools:domain_dns_configuration", args=[self.ac_rec.domain.pk])
        response = self.client.get(url)
        self.assertContains(response, "[IP address of your Modoboa server]")


class DomainTestCase(ModoTestCase):

    def test_dns_global_status(self):
        admin_factories.MXRecordFactory.create(
            domain__name="test.com", name="mail.test.com",
            updated=timezone.now())
        domain = admin_models.Domain.objects.get(name="test.com")
        self.assertEqual(domain.dns_global_status, "critical")

        factories.DNSRecordFactory.create(
            type="spf", value="v=SPF1 mx -all", is_valid=True,
            domain__name="test.com"
        )
        self.assertEqual(domain.dns_global_status, "critical")

        factories.DNSRecordFactory(
            type="dkim", value="12345", is_valid=True,
            domain__name="test.com"
        )
        self.assertEqual(domain.dns_global_status, "critical")

        factories.DNSRecordFactory(
            type="dmarc", value="XXX", is_valid=True,
            domain__name="test.com"
        )
        self.assertEqual(domain.dns_global_status, "critical")

        factories.DNSRecordFactory(
            type="autoconfig", value="1.2.3.4", is_valid=True,
            domain__name="test.com"
        )
        self.assertEqual(domain.dns_global_status, "critical")

        factories.DNSRecordFactory(
            type="autodiscover", value="1.2.3.4", is_valid=True,
            domain__name="test.com"
        )
        self.assertEqual(domain.dns_global_status, "ok")



import email
import re
import smtplib
import time
from email.header import Header
from email.mime.text import MIMEText
from email.utils import formataddr, formatdate, getaddresses, make_msgid

import chardet
import lxml.html
from lxml.html import defs
from lxml.html.clean import Cleaner

from django.core.mail import EmailMessage
from django.template.loader import render_to_string
from django.utils.encoding import smart_str, smart_text
from django.utils.html import conditional_escape, escape
from django.utils.translation import ugettext as _

from modoboa.core import models as core_models
from modoboa.lib import u2u_decode
from modoboa.lib.exceptions import InternalError

# used by Email()
_RE_REMOVE_EXTRA_WHITESPACE = re.compile(r"\n\s*\n")
_RE_CID = re.compile(r".*[ ]*cid=\"([^\"]*)\".*", re.I)


class EmailAddress(object):

    def __init__(self, address):
        self.name, self.address = u2u_decode.decode_address(address)
        if self.name:
            self.fulladdress = "{} <{}>".format(self.name, self.address)
        else:
            self.fulladdress = self.address

    def __str__(self):
        return self.fulladdress


class Email(object):

    _basic_headers = (
        "From",
        "To",
        "Cc",
        "Date",
        "Subject",
    )

    _image_mime_types = (
        "image/png",
        "image/gif",
        "image/jpeg",
    )

    def __init__(self, mailid, mformat="plain", dformat="plain", links=False):
        self.contents = {"html": "", "plain": ""}
        self.mailid = mailid
        self.mformat = mformat
        self.dformat = dformat

        if isinstance(links, bool):
            self.links = links
        elif isinstance(links, int):
            # TODO: modoboa-webmail currently sets links == 0 or 1, it needs
            #       updated to use True or False.
            self.links = bool(links)
        elif links == "0":
            raise TypeError("links == \"0\" is not valid, did you mean True or "
                            "False?")
        else:
            raise TypeError("links should be a boolean value")

        self._msg = None
        self._headers = None
        self._body = None
        self._images = {}

        # used by modoboa_webmail.lib.imapemail.ImapEmail
        self.attachments = {}

    @property
    def msg(self):
        if self._msg is None:
            mail_text = self._fetch_message()
            # python 2 expects str (bytes), python 3 expects str (unicode)
            mail_text = smart_str(mail_text, errors="replace")
            self._msg = email.message_from_string(mail_text)

        return self._msg

    @property
    def headers(self):
        if self._headers is None:
            self._headers = []

            # don't break modoboa_webmail.lib.imapemail.ImapEmail, it
            # doesn't have self._msg it loads e-mails (and their headers)
            # from the IMAP server.
            if self._msg is not None:
                for header in self._basic_headers:
                    value = self.get_header(self.msg, header)
                    self._headers.append({
                        "name": header,
                        "value": value
                    })

        return self._headers

    @headers.setter
    def headers(self, value):
        self._headers = value

    @property
    def body(self):
        if self._body is None:
            self._parse(self.msg)
            self._body = getattr(self, "viewmail_%s" % self.dformat)()

        return self._body

    def get_header(self, msg, header):
        # msg parameter to maintain compatibility with
        # modoboa_webmail.lib.imapemail.ImapEmail
        if header in msg:
            decoded_values = []
            for value, encoding in email.header.decode_header(msg[header]):
                if not len(value):
                    continue
                if encoding:
                    value = smart_text(value, encoding=encoding)
                elif isinstance(value, bytes):
                    # SMTPUTF8 fallback (most of the time)
                    # Address contains non ASCII chars but is not RFC2047
                    # encoded...
                    encoding = chardet.detect(value)
                    try:
                        value = value.decode(encoding["encoding"], "replace")
                    except (TypeError, UnicodeDecodeError) as exc:
                        raise InternalError(
                            _("unable to determine encoding of string")
                        ) from exc
                decoded_values.append(value)
            return "".join(decoded_values)
        return ""

    def _fetch_message(self):
        raise NotImplementedError()

    def _parse(self, msg, level=0):
        """Recursive email parser

        A message structure can be complex. To correctly handle
        unknown MIME types, a simple rule is applied : if I don't know
        how to display a specific part, it becomes an attachment! If
        no name is specified for an attachment, the part number
        described in the RFC 3501 (which retrieves BODY sections) is
        used to build a file name (like part_1.1).

        :param msg: message (or part) to parse
        :param level: current part number
        """
        content_type = msg.get_content_maintype()
        is_attachment = "attachment" in msg.get("Content-Disposition", "")

        if content_type == "text" and not is_attachment:
            self._parse_text(msg, level)
        elif content_type == "multipart" and not is_attachment:
            self._parse_multipart(msg, level)
        else:
            self._parse_inline_image(msg, level)

        if self.contents["plain"]:
            self.contents["plain"] = self._post_process_plain(
                self.contents["plain"])
        if self.contents["html"]:
            self.contents["html"] = self._post_process_html(
                self.contents["html"])

    def _parse_text(self, msg, level=0):
        content_type = msg.get_content_subtype()

        if content_type in ["plain", "html"]:
            encoding = msg.get_content_charset() or "utf-8"
            mail_text_bytes = msg.get_payload(decode=True)
            mail_text = decode(mail_text_bytes, encoding)
            self.contents[content_type] += mail_text

    def _parse_multipart(self, msg, level=0):
        level += 1
        for part in msg.walk():
            content_type = part.get_content_maintype()
            if content_type == "text":
                self._parse_text(part, level=level)
            else:
                # I'm  a dumb mail parser and treat all non-text parts as
                # attachments
                self._parse_inline_image(part, level=level)

    def _parse_inline_image(self, msg, level=0):
        content_type = msg.get_content_type()
        cid = None
        if "Content-ID" in msg:
            cid = msg["Content-ID"]
            if cid.startswith("<") and cid.endswith(">"):
                cid = cid[1:-1]
        else:
            matches = _RE_CID.match(msg["Content-Type"])
            if matches is not None:
                cid = matches.group(1)
        transfer_encoding = msg["Content-Transfer-Encoding"]
        if content_type not in self._image_mime_types or cid is None:
            # I'm a dumb mail parser and I don't care ;p
            return
        if transfer_encoding != "base64":
            # I don't deal with non base64 encoded images
            return
        if cid in self._images:
            # Duplicate Content-ID
            return

        self._images[cid] = "data:%s;base64,%s" % (
            content_type, "".join(msg.get_payload().splitlines(False)))

    def _post_process_plain(self, content):
        mail_text = (
            _RE_REMOVE_EXTRA_WHITESPACE.sub("\n\n", content).strip()
        )
        mail_text = escape(mail_text)
        return smart_text(mail_text)

    def _post_process_html(self, content):
        html = lxml.html.fromstring(content)
        if self.links:
            html.rewrite_links(self._map_cid)

            for link in html.iterlinks():
                link[0].set("target", "_blank")
        else:
            html.rewrite_links(lambda x: None)
        safe_attrs = list(defs.safe_attrs) + ["class", "style"]
        cleaner = Cleaner(
            scripts=True,
            javascript=True,
            links=True,
            page_structure=True,
            embedded=True,
            frames=True,
            add_nofollow=True,
            safe_attrs=safe_attrs
        )
        mail_text = lxml.html.tostring(
            cleaner.clean_html(html), encoding="unicode")
        with open("/tmp/output.txt", "w") as fp:
            fp.write(mail_text)
        return smart_text(mail_text)

    def _map_cid(self, url):
        if url.startswith("cid:"):
            cid = url[4:]
            if cid in self._images:
                url = self._images[cid]

        return url

    def viewmail_plain(self, contents=None, **kwargs):
        # contents and **kwargs parameters to maintain compatibility with
        # modoboa_webmail.lib.imapemail.ImapEmail
        if contents is None:
            contents = self.contents["plain"]
        else:
            contents = conditional_escape(contents)
        return "<pre>%s</pre>" % contents

    def viewmail_html(self, contents=None, **kwargs):
        # contents and **kwargs parameters to maintain compatibility with
        # modoboa_webmail.lib.imapemail.ImapEmail
        if contents is None:
            contents = self.contents["html"]
        return contents

    def render_headers(self, **kwargs):
        context = {
            "headers": self.headers,
        }
        return render_to_string("common/mailheaders.html", context)


def split_address(address):
    """Split an e-mail address into local part and domain."""
    assert isinstance(address, str),\
        "address should be of type str"
    if "@" not in address:
        local_part = address
        domain = None
    else:
        local_part, domain = address.rsplit("@", 1)
    return (local_part, domain)


def split_local_part(local_part, delimiter=None):
    """Split a local part into local part and extension."""
    assert isinstance(local_part, str),\
        "local_part should be of type str"
    assert isinstance(delimiter, str) or delimiter is None,\
        "delimiter should be of type str"
    extension = None
    if local_part.lower() in ["mailer-daemon", "double-bounce"]:
        # never split these special addresses
        pass
    elif (
        delimiter == "-" and
         (local_part.startswith("owner-") or
          local_part.endswith("-request"))
    ):
        # don't split owner-* or *-request if - is the delimiter, they are
        # special addresses used by mail lists.
        pass
    elif (
        delimiter and
        delimiter in local_part and
        local_part[0] != delimiter and
        local_part[-1] != delimiter
    ):
        local_part, extension = local_part.split(delimiter, 1)

    return (local_part, extension)


def split_mailbox(mailbox, return_extension=False):
    """Try to split an address into parts (local part and domain name).

    If return_extension is True, we also look for an address extension
    (something foo+bar).

    :return: a tuple (local part, domain<, extension>)

    """
    local_part, domain = split_address(mailbox)
    if not return_extension:
        return (local_part, domain)
    local_part, extension = split_local_part(local_part, delimiter="+")
    return (local_part, domain, extension)


def decode(value_bytes, encoding, append_to_error=""):
    """Try to decode the given string."""
    assert isinstance(value_bytes, bytes),\
        "value_bytes should be of type bytes"
    if len(value_bytes) == 0:
        # short circuit for empty strings
        return ""
    try:
        value = value_bytes.decode(encoding)
    except (UnicodeDecodeError, LookupError):
        encoding = chardet.detect(value_bytes)
        try:
            value = value_bytes.decode(
                encoding["encoding"], "replace"
            )
        except (TypeError, UnicodeDecodeError) as exc:
            raise InternalError(
                _("unable to determine encoding of string") +
                append_to_error
            ) from exc
    return value


def prepare_addresses(addresses, usage="header"):
    """Prepare addresses before using them

    :param list addresses: a list of addresses
    :param string usage: how those addresses are going to be used
    :return: a string or a list depending on usage value
    """
    result = []
    if isinstance(addresses, str):
        addresses = [addresses]
    for name, address in getaddresses(addresses):
        if not address:
            continue
        if name and usage == "header":
            result.append(formataddr((name, address)))
        else:
            result.append(address)
    if usage == "header":
        return ",".join(result)
    return result


def set_email_headers(msg, subject, sender, rcpt):
    """Set mandatory headers.

    Subject, From, To, Message-ID, User-Agent, Date
    """
    import pkg_resources

    msg["Subject"] = Header(subject, "utf8")
    msg["From"] = sender
    msg["To"] = prepare_addresses(rcpt)
    msg["Message-ID"] = make_msgid()
    msg["User-Agent"] = "Modoboa %s" % \
        (pkg_resources.get_distribution("modoboa").version)
    msg["Date"] = formatdate(time.time(), True)


def _sendmail(sender, rcpt, msgstring, server="localhost", port=25):
    """Message sending

    Return a tuple (True, None) on success, (False, error message)
    otherwise.

    :param sender: sender address
    :param rcpt: recipient address
    :param msgstring: the message structure (must be a string)
    :param server: the sending server's address
    :param port: the listening port
    :return: tuple
    """
    try:
        s = smtplib.SMTP(server, port)
        s.sendmail(sender, [rcpt], msgstring)
        s.quit()
    except smtplib.SMTPException as e:
        return False, "SMTP error: %s" % str(e)
    return True, None


def sendmail_simple(
        sender, rcpt, subject="Sample message", content="", **kwargs):
    """Simple way to send a text message

    Send a text/plain message with basic headers (msg-id, date).

    Return a tuple (True, None) on success, (False, error message)
    otherwise.

    :param sender: sender address
    :param rcpt: recipient address
    :param subject: message's subject
    :param content: message's content
    :return: tuple
    """
    msg = MIMEText(content, _charset="utf-8")
    set_email_headers(msg, subject, sender, rcpt)
    return _sendmail(sender, rcpt, msg.as_string(), **kwargs)


def sendmail_fromfile(sender, rcpt, fname):
    """Send a message contained within a file

    The given file name must represent a valid message structure. It
    must not include the From: and To: headers, they are automatically
    added by the function.

    :param sender: sender address
    :param rcpt: recipient address
    :param fname: the name of the file containing the message
    :return: a tuple
    """
    try:
        fp = open(fname)
    except IOError as e:
        return False, str(e)

    content = """From: %s
To: %s
""" % (sender, rcpt)
    content += fp.read()
    fp.close()

    return _sendmail(sender, rcpt, content)


def send_notification(recipient, subject, tpl, sender=None, **kwargs):
    """Send notification by email."""
    if not sender:
        local_config = (
            core_models.LocalConfig.objects.select_related("site").first())
        sender = local_config.parameters.get_value("sender_address")
    content = render_to_string(tpl, kwargs)
    msg = EmailMessage(
        subject, content.strip(), sender, [recipient],
    )
    msg.send()



from django.urls import reverse

from modoboa.lib.tests import ModoTestCase
from .. import factories
from ..models import Alias, Domain, DomainAlias


class DomainAliasTestCase(ModoTestCase):

    @classmethod
    def setUpTestData(cls):  # NOQA:N802
        """Create test data."""
        super(DomainAliasTestCase, cls).setUpTestData()
        factories.populate_database()
        cls.dom = Domain.objects.get(name="test.com")

    def test_model(self):
        dom = Domain.objects.get(name="test.com")
        domal = DomainAlias()
        domal.name = "domalias.net"
        domal.target = dom
        domal.save()
        self.assertEqual(dom.domainalias_count, 1)
        self.assertTrue(
            Alias.objects.filter(
                address="@{}".format(domal.name)).exists())

        domal.name = "domalias.org"
        domal.save()

        domal.delete()
        self.assertFalse(
            Alias.objects.filter(
                address="@{}".format(domal.name)).exists())

    def test_form(self):
        dom = Domain.objects.get(name="test.com")
        values = {
            "name": dom.name, "quota": dom.quota,
            "default_mailbox_quota": dom.default_mailbox_quota,
            "enabled": dom.enabled, "aliases": "domalias.net",
            "aliases_1": "domaliasé.com", "type": "domain"
        }
        self.ajax_post(
            reverse("admin:domain_change",
                    args=[dom.id]),
            values
        )
        self.assertEqual(dom.domainalias_set.count(), 2)

        del values["aliases_1"]
        self.ajax_post(
            reverse("admin:domain_change",
                    args=[dom.id]),
            values
        )
        self.assertEqual(dom.domainalias_set.count(), 1)



"""Test cases for the limits extension."""

from django.urls import reverse

from modoboa.admin.factories import populate_database
from modoboa.admin.models import Alias, Domain
from modoboa.core.factories import UserFactory
from modoboa.core.models import User
from modoboa.lib import tests as lib_tests
from .. import utils


class PermissionsTestCase(lib_tests.ModoTestCase):

    @classmethod
    def setUpTestData(cls):  # NOQA:N802
        """Create test data."""
        super(PermissionsTestCase, cls).setUpTestData()
        populate_database()

    def test_domainadmin_deletes_reseller(self):
        """Check if a domain admin can delete a reseller.

        Expected result: no.
        """
        values = {
            "username": "reseller@test.com", "first_name": "Reseller",
            "last_name": "", "password1": "Toto1234", "password2": "Toto1234",
            "role": "Resellers", "is_active": True,
            "email": "reseller@test.com", "stepid": "step2"
        }
        self.ajax_post(reverse("admin:account_add"), values)
        account = User.objects.get(username="reseller@test.com")
        admin = User.objects.get(username="admin@test.com")
        self.client.force_login(admin)
        resp = self.ajax_post(
            reverse("admin:account_delete", args=[account.id]),
            {}, status=403
        )
        self.assertEqual(resp, "Permission denied")


class ResourceTestCase(lib_tests.ModoTestCase):

    @classmethod
    def setUpTestData(cls):  # NOQA:N802
        """Custom setUpTestData method."""
        super(ResourceTestCase, cls).setUpTestData()
        cls.localconfig.parameters.set_values({
            "enable_admin_limits": True,
            "enable_domain_limits": False
        })
        for name, _definition in utils.get_user_limit_templates():
            cls.localconfig.parameters.set_value(
                "deflt_user_{0}_limit".format(name), 2)
        cls.localconfig.save()
        populate_database()

    def _create_account(
            self, username, role="SimpleUsers", status=200, **kwargs):
        values = {
            "username": username, "first_name": "Tester", "last_name": "Toto",
            "password1": "Toto1234", "password2": "Toto1234", "role": role,
            "quota_act": True, "is_active": True, "email": username,
            "stepid": "step2"
        }
        values.update(kwargs)
        return self.ajax_post(
            reverse("admin:account_add"), values, status
        )

    def _create_alias(self, email, rcpt="user@test.com", status=200):
        values = {
            "address": email, "recipients": rcpt, "enabled": True
        }
        return self.ajax_post(
            reverse("admin:alias_add"), values, status
        )

    def _create_domain(self, name, status=200, withtpl=False, **kwargs):
        values = {
            "name": name, "quota": 100, "default_mailbox_quota": 10,
            "create_dom_admin": False, "with_mailbox": True,
            "create_aliases": False, "stepid": "step3", "type": "domain"
        }
        if withtpl:
            values.update({
                "create_dom_admin": True,
                "dom_admin_username": "admin",
                "create_aliases": True
            })
        values.update(kwargs)
        response = self.ajax_post(
            reverse("admin:domain_add"), values, status
        )
        return response

    def _domain_alias_operation(self, optype, domain, name, status=200):
        dom = Domain.objects.get(name=domain)
        values = {
            "name": dom.name, "quota": dom.quota, "enabled": dom.enabled,
            "type": "domain",
            "default_mailbox_quota": dom.default_mailbox_quota,
        }
        aliases = [alias.name for alias in dom.domainalias_set.all()]
        if optype == "add":
            aliases.append(name)
        else:
            aliases.remove(name)
        for cpt, alias in enumerate(aliases):
            fname = "aliases" if not cpt else "aliases_%d" % cpt
            values[fname] = alias
        self.ajax_post(
            reverse("admin:domain_change", args=[dom.id]),
            values, status
        )

    def _check_limit(self, name, curvalue, maxvalue):
        limit = self.user.userobjectlimit_set.get(name=name)
        self.assertEqual(limit.current_value, curvalue)
        self.assertEqual(limit.max_value, maxvalue)


class DomainAdminTestCase(ResourceTestCase):

    @classmethod
    def setUpTestData(cls):  # NOQA:N802
        """Create test data."""
        super(DomainAdminTestCase, cls).setUpTestData()
        cls.user = User.objects.get(username="admin@test.com")
        cls.user.userobjectlimit_set.filter(
            name__in=["mailboxes", "mailbox_aliases"]).update(max_value=2)

    def setUp(self):
        """Test initialization."""
        super(DomainAdminTestCase, self).setUp()
        self.client.force_login(self.user)

    def test_mailboxes_limit(self):
        self._create_account("tester1@test.com")
        self._check_limit("mailboxes", 1, 2)
        self._create_account("tester2@test.com")
        self._check_limit("mailboxes", 2, 2)
        self._create_account("tester3@test.com", status=403)
        self._check_limit("mailboxes", 2, 2)
        self.ajax_post(
            reverse("admin:account_delete",
                    args=[User.objects.get(username="tester2@test.com").id]),
            {}
        )
        self._check_limit("mailboxes", 1, 2)

    def test_aliases_limit(self):
        self._create_alias("alias1@test.com")
        self._check_limit("mailbox_aliases", 1, 2)
        self._create_alias("alias2@test.com")
        self._check_limit("mailbox_aliases", 2, 2)
        self._create_alias("alias3@test.com", status=403)
        self._check_limit("mailbox_aliases", 2, 2)
        # Set unlimited value
        self.user.userobjectlimit_set.filter(name="mailbox_aliases").update(
            max_value=-1)
        self._create_alias("alias3@test.com")
        self._check_limit("mailbox_aliases", 3, -1)
        self.ajax_post(
            reverse("admin:alias_delete") + "?selection=%d"
            % Alias.objects.get(address="alias2@test.com").id,
            {}
        )
        self._check_limit("mailbox_aliases", 2, -1)

    def test_aliases_limit_through_account_form(self):
        user = User.objects.get(username="user@test.com")
        values = {
            "username": user.username, "role": user.role,
            "is_active": user.is_active, "email": user.email, "quota_act": True,
            "aliases": "alias1@test.com", "aliases_1": "alias2@test.com",
            "language": "en"
        }
        self.ajax_post(
            reverse("admin:account_change", args=[user.id]),
            values
        )
        Alias.objects.get(address="alias1@test.com")
        self._check_limit("mailbox_aliases", 2, 2)


class ResellerTestCase(ResourceTestCase):

    @classmethod
    def setUpTestData(cls):  # NOQA:N802
        """Create test data."""
        super(ResellerTestCase, cls).setUpTestData()
        cls.localconfig.parameters.set_value("deflt_user_quota_limit", 1000)
        cls.localconfig.save()
        cls.user = UserFactory(
            username="reseller", groups=("Resellers",)
        )

    def setUp(self):
        """Test initialization."""
        super(ResellerTestCase, self).setUp()
        self.client.force_login(self.user)

    def test_domains_limit(self):
        response = self.client.get(reverse("admin:domain_list"))
        self.assertContains(response, "Domains (0%)")
        self.assertContains(response, "Domain aliases (0%)")
        self._create_domain("domain1.tld")
        self._check_limit("domains", 1, 2)
        self._create_domain("domain2.tld")
        self._check_limit("domains", 2, 2)
        self._create_domain("domain3.tld", 403)
        self._check_limit("domains", 2, 2)
        self.ajax_post(
            reverse("admin:domain_delete",
                    args=[Domain.objects.get(name="domain2.tld").id]),
            {}
        )
        self._check_limit("domains", 1, 2)

    def test_domain_aliases_limit(self):
        self._create_domain("pouet.com")
        self._domain_alias_operation("add", "pouet.com", "domain-alias1.tld")
        self._check_limit("domain_aliases", 1, 2)
        self._domain_alias_operation("add", "pouet.com", "domain-alias2.tld")
        self._check_limit("domain_aliases", 2, 2)
        self._domain_alias_operation(
            "add", "pouet.com", "domain-alias3.tld", 403
        )
        self._check_limit("domain_aliases", 2, 2)
        self._domain_alias_operation(
            "delete", "pouet.com", "domain-alias2.tld")
        self._check_limit("domain_aliases", 1, 2)

    def test_domain_admins_limit(self):
        response = self.client.get(reverse("admin:identity_list"))
        self.assertContains(response, "Domain admins (0%)")
        self.assertContains(response, "Mailboxes (0%)")
        self.assertContains(response, "Mailbox aliases (0%)")

        self._create_domain("domain.tld")
        self._create_account("admin1@domain.tld", role="DomainAdmins")
        self._check_limit("domain_admins", 1, 2)
        self._create_account("admin2@domain.tld", role="DomainAdmins")
        self._check_limit("domain_admins", 2, 2)
        resp = self._create_account(
            "admin3@domain.tld",
            role="DomainAdmins",
            status=400)
        self.assertEqual(
            resp["form_errors"]["role"][0],
            "Select a valid choice. DomainAdmins is not one of the available "
            "choices."
        )
        self._check_limit("domain_admins", 2, 2)

        self.user.userobjectlimit_set.filter(
            name="mailboxes").update(max_value=3)
        self._create_account("user1@domain.tld")
        user = User.objects.get(username="user1@domain.tld")
        values = {
            "username": user.username, "role": "DomainAdmins",
            "quota_act": True, "is_active": user.is_active,
            "email": user.email, "language": "en"
        }
        resp = self.ajax_post(
            reverse("admin:account_change", args=[user.id]),
            values, status=400
        )
        self.assertEqual(
            resp["form_errors"]["role"][0],
            "Select a valid choice. DomainAdmins is not one of the available "
            "choices."
        )
        self._check_limit("domain_admins", 2, 2)

    def test_domain_admin_resource_are_empty(self):
        self._create_domain("domain.tld")
        self._create_account("admin1@domain.tld", role="DomainAdmins")
        domadmin = User.objects.get(username="admin1@domain.tld")
        for l in ["mailboxes", "mailbox_aliases"]:
            self.assertEqual(
                domadmin.userobjectlimit_set.get(name=l).max_value, 0
            )

    def test_domain_admins_limit_from_domain_tpl(self):
        self.user.userobjectlimit_set.filter(
            name="domains").update(max_value=3)
        self._create_domain("domain1.tld", withtpl=True)
        self._create_domain("domain2.tld", withtpl=True)
        self._check_limit("domain_admins", 2, 2)
        self._check_limit("domains", 2, 3)
        self._create_domain("domain3.tld", status=200, withtpl=True)
        self._check_limit("domain_admins", 2, 2)
        self._check_limit("domains", 3, 3)

    def test_quota(self):
        """Check quota resource."""
        self._create_domain("domain1.tld", withtpl=True, quota=1000)
        response = self._create_domain(
            "domain2.tld", status=403, withtpl=True, quota=1000)
        self.assertEqual(response, "Quota: limit reached")
        dom1 = Domain.objects.get(name="domain1.tld")
        url = reverse("admin:domain_change", args=[dom1.pk])
        values = {
            "name": dom1.name, "type": dom1.type, "enabled": dom1.enabled,
            "default_mailbox_quota": dom1.default_mailbox_quota,
            "quota": 500
        }
        self.ajax_post(url, values)

    def test_quota_constraints(self):
        """Check reseller can't define unlimited quota."""
        response = self._create_domain("domain1.tld", 400, quota=0)
        self.assertEqual(
            response["form_errors"]["quota"][0],
            "You can't define an unlimited quota.")
        response = self._create_domain(
            "domain1.tld", 400, default_mailbox_quota=0)
        self.assertEqual(
            response["form_errors"]["default_mailbox_quota"][0],
            "You can't define an unlimited quota.")

        self.user.userobjectlimit_set.filter(name="quota").update(max_value=0)
        response = self._create_domain("domain2.tld", quota=0)

    def test_quota_propagation(self):
        """Check that quota is applied everywhere."""
        # Try to assign an unlimited quota to an admin
        # See https://github.com/modoboa/modoboa/issues/1223
        self._create_domain("domain1.tld")
        self._create_account(
            "admin@domain1.tld", role="DomainAdmins",
            quota_act=False, quota=0, status=400)

        self._create_domain("domain2.tld", withtpl=True)
        admin = User.objects.get(username="admin@domain2.tld")
        values = {
            "username": admin.username, "role": admin.role,
            "is_active": admin.is_active, "email": admin.email,
            "mailboxes_limit": 1, "mailbox_aliases_limit": 2,
            "language": "en", "quota_act": False, "quota": 0
        }
        self.ajax_post(
            reverse("admin:account_change", args=[admin.pk]),
            values,
            400
        )

    def test_reseller_deletes_domain(self):
        """Check if all resources are restored after the deletion."""
        self._create_domain("domain.tld", withtpl=True)
        dom = Domain.objects.get(name="domain.tld")
        self.ajax_post(
            reverse("admin:domain_delete", args=[dom.id]),
            {}
        )
        self._check_limit("domains", 0, 2)
        self._check_limit("domain_admins", 1, 2)
        self._check_limit("mailboxes", 0, 2)
        self._check_limit("mailbox_aliases", 0, 2)

    def test_sadmin_removes_ownership(self):
        self._create_domain("domain.tld", withtpl=True)
        dom = Domain.objects.get(name="domain.tld")
        self.client.logout()
        self.client.login(username="admin", password="password")
        self.ajax_get(
            "{0}?domid={1}&daid={2}".format(
                reverse("admin:permission_remove"),
                dom.id, self.user.id
            ), {}
        )
        self._check_limit("domains", 0, 2)
        self._check_limit("domain_admins", 0, 2)
        self._check_limit("mailboxes", 0, 2)
        self._check_limit("mailbox_aliases", 0, 2)

    def test_allocate_from_pool(self):
        self._create_domain("domain.tld")
        self._create_account("admin1@domain.tld", role="DomainAdmins")
        user = User.objects.get(username="admin1@domain.tld")

        # Give 1 mailbox and 2 aliases to the admin -> should work
        values = {
            "username": user.username, "role": user.role, "quota_act": True,
            "is_active": user.is_active, "email": user.email,
            "mailboxes_limit": 1, "mailbox_aliases_limit": 2,
            "language": "en"
        }
        self.ajax_post(
            reverse("admin:account_change", args=[user.id]),
            values
        )
        self._check_limit("mailboxes", 1, 1)
        self._check_limit("mailbox_aliases", 0, 0)

        # Delete the admin -> resources should go back to the
        # reseller's pool
        self.ajax_post(
            reverse("admin:account_delete", args=[user.id]),
            {}
        )
        self._check_limit("mailboxes", 0, 2)
        self._check_limit("mailbox_aliases", 0, 2)

    def test_restore_resources(self):
        """Give resource to a domain admin and restore them."""
        self._create_domain("domain.tld")
        dom = Domain.objects.get(name="domain.tld")
        self._create_account("admin1@domain.tld", role="DomainAdmins")
        user = User.objects.get(username="admin1@domain.tld")
        values = {
            "username": user.username, "role": user.role, "quota_act": True,
            "is_active": user.is_active, "email": user.email,
            "mailboxes_limit": 1, "mailbox_aliases_limit": 2,
            "language": "en"
        }
        self.ajax_post(
            reverse("admin:account_change", args=[user.id]),
            values
        )
        dom.add_admin(user)
        self.client.logout()
        self.client.login(username="admin1@domain.tld", password="Toto1234")
        self._create_account("user1@domain.tld")
        self._create_alias("alias1@domain.tld", "user1@domain.tld")
        self._create_alias("alias2@domain.tld", "user1@domain.tld")
        self.client.logout()
        self.client.login(username="reseller", password="toto")
        # Delete the admin -> resources should go back to the
        # reseller's pool
        self.ajax_post(
            reverse("admin:account_delete", args=[user.id]),
            {}
        )
        self._check_limit("mailboxes", 1, 2)
        self._check_limit("mailbox_aliases", 2, 2)

    def test_change_role(self):
        self._create_domain("domain.tld")
        self._create_account("admin1@domain.tld", role="DomainAdmins")
        user = User.objects.get(username="admin1@domain.tld")

        # Give 1 mailbox and 2 aliases to the admin -> should work
        values = {
            "username": user.username, "role": user.role, "quota_act": True,
            "is_active": user.is_active, "email": user.email,
            "mailboxes_limit": 1, "mailbox_aliases_limit": 2,
            "language": "en"
        }
        self.ajax_post(
            reverse("admin:account_change", args=[user.id]),
            values
        )
        self._check_limit("mailboxes", 1, 1)
        self._check_limit("mailbox_aliases", 0, 0)

        # Change admin role to SimpleUser -> resources should go back
        # to the reseller.
        values = {
            "username": user.username, "role": "SimpleUsers",
            "quota_act": True,
            "is_active": user.is_active, "email": user.email,
            "language": "en",
        }
        self.ajax_post(
            reverse("admin:account_change", args=[user.id]),
            values
        )
        self._check_limit("mailboxes", 1, 2)
        self._check_limit("mailbox_aliases", 0, 2)

    def test_allocate_too_much(self):
        self._create_domain("domain.tld")
        self._create_account("admin1@domain.tld", role="DomainAdmins")
        user = User.objects.get(username="admin1@domain.tld")

        # Give 2 mailboxes and 3 aliases to the admin -> should fail.
        values = {
            "username": user.username, "role": user.role, "quota_act": True,
            "is_active": user.is_active, "email": user.email,
            "mailboxes_limit": 2, "mailbox_aliases_limit": 3,
            "language": "en"
        }
        resp = self.ajax_post(
            reverse("admin:account_change", args=[user.id]),
            values, 424
        )
        self.assertEqual(resp, "Not enough resources")
        self._check_limit("mailboxes", 1, 2)
        self._check_limit("mailbox_aliases", 0, 2)



"""Core API v2 serializers."""

from django.utils import formats
from django.utils.translation import ugettext_lazy, ugettext as _

from django.contrib.auth import password_validation

import django_otp
from rest_framework import serializers

from modoboa.lib import fields as lib_fields

from ... import constants
from ... import models
from ... import sms_backends


class CoreGlobalParametersSerializer(serializers.Serializer):
    """A serializer for global parameters."""

    # General settings
    authentication_type = serializers.ChoiceField(
        choices=[("local", ugettext_lazy("Local")),
                 ("ldap", "LDAP")],
        default="local"
    )
    password_scheme = serializers.ChoiceField(
        choices=[("sha512crypt", "sha512crypt"),
                 ("sha256crypt", "sha256crypt"),
                 ("blfcrypt", "bcrypt"),
                 ("md5crypt", ugettext_lazy("md5crypt (weak)")),
                 ("sha256", ugettext_lazy("sha256 (weak)")),
                 ("md5", ugettext_lazy("md5 (weak)")),
                 ("crypt", ugettext_lazy("crypt (weak)")),
                 ("plain", ugettext_lazy("plain (weak)"))],
        default="sha512crypt"
    )
    rounds_number = serializers.IntegerField(default=70000)
    update_scheme = serializers.BooleanField(default=True)
    default_password = serializers.CharField(default="password")
    random_password_length = serializers.IntegerField(min_value=8, default=8)
    update_password_url = serializers.URLField(required=False, allow_blank=True)
    password_recovery_msg = serializers.CharField(
        required=False, allow_blank=True)
    sms_password_recovery = serializers.BooleanField(default=False)
    sms_provider = serializers.ChoiceField(
        choices=constants.SMS_BACKENDS, required=False)

    # LDAP settings
    ldap_server_address = serializers.CharField(default="localhost")
    ldap_server_port = serializers.IntegerField(default=389)
    ldap_enable_secondary_server = serializers.BooleanField(default=False)
    ldap_secondary_server_address = serializers.CharField(required=False)
    ldap_secondary_server_port = serializers.IntegerField(
        default=389, required=False)
    ldap_secured = serializers.ChoiceField(
        choices=constants.LDAP_SECURE_MODES,
        default="none"
    )
    ldap_is_active_directory = serializers.BooleanField(default=False)
    ldap_admin_groups = serializers.CharField(
        default="", required=False, allow_blank=True)
    ldap_group_type = serializers.ChoiceField(
        default="posixgroup",
        choices=constants.LDAP_GROUP_TYPES
    )
    ldap_groups_search_base = serializers.CharField(
        default="", required=False, allow_blank=True)
    ldap_password_attribute = serializers.CharField(default="userPassword")

    # LDAP auth settings
    ldap_auth_method = serializers.ChoiceField(
        choices=constants.LDAP_AUTH_METHODS,
        default="searchbind",
    )
    ldap_bind_dn = serializers.CharField(
        default="", required=False, allow_blank=True)
    ldap_bind_password = serializers.CharField(
        default="", required=False, allow_blank=True)
    ldap_search_base = serializers.CharField(
        default="", required=False, allow_blank=True)
    ldap_search_filter = serializers.CharField(
        default="(mail=%(user)s)", required=False,
        allow_blank=True)
    ldap_user_dn_template = serializers.CharField(
        default="", required=False, allow_blank=True)

    # LDAP sync settings
    ldap_sync_bind_dn = serializers.CharField(required=False, allow_blank=True)
    ldap_sync_bind_password = serializers.CharField(
        required=False, allow_blank=True)
    ldap_enable_sync = serializers.BooleanField(default=False)
    ldap_sync_delete_remote_account = serializers.BooleanField(default=False)
    ldap_sync_account_dn_template = serializers.CharField(
        required=False, allow_blank=True)
    ldap_enable_import = serializers.BooleanField(default=False)
    ldap_import_search_base = serializers.CharField(
        required=False, allow_blank=True)
    ldap_import_search_filter = serializers.CharField(
        default="(cn=*)", required=False
    )
    ldap_import_username_attr = serializers.CharField(default="cn")
    ldap_dovecot_sync = serializers.BooleanField(default=False)
    ldap_dovecot_conf_file = serializers.CharField(
        default="/etc/dovecot/dovecot-modoboa.conf", required=False
    )

    # Dashboard settings
    rss_feed_url = serializers.URLField(allow_blank=True)
    hide_features_widget = serializers.BooleanField(default=False)

    # Notification settings
    sender_address = lib_fields.DRFEmailFieldUTF8(
        default="noreply@yourdomain.test")

    # API settings
    enable_api_communication = serializers.BooleanField(default=True)
    check_new_versions = serializers.BooleanField(default=True)
    send_new_versions_email = serializers.BooleanField(default=False)
    new_versions_email_rcpt = lib_fields.DRFEmailFieldUTF8(required=False)
    send_statistics = serializers.BooleanField(default=True)

    # Misc settings
    inactive_account_threshold = serializers.IntegerField(default=30)
    top_notifications_check_interval = serializers.IntegerField(default=30)
    log_maximum_age = serializers.IntegerField(default=365)
    items_per_page = serializers.IntegerField(default=30)
    default_top_redirection = serializers.ChoiceField(
        default="user", choices=[""])

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        sms_backend_fields = sms_backends.get_all_backend_serializer_settings()
        for field, definition in sms_backend_fields.items():
            self.fields[field] = definition["type"](
                **definition["attrs"])

    def validate_ldap_user_dn_template(self, value):
        try:
            value % {"user": "toto"}
        except (KeyError, ValueError):
            raise serializers.ValidationError(_("Invalid syntax"))
        return value

    def validate_ldap_sync_account_dn_template(self, value):
        try:
            value % {"user": "toto"}
        except (KeyError, ValueError):
            raise serializers.ValidationError(_("Invalid syntax"))
        return value

    def validate_ldap_search_filter(self, value):
        try:
            value % {"user": "toto"}
        except (KeyError, ValueError, TypeError):
            raise serializers.ValidationError(_("Invalid syntax"))
        return value

    def validate_rounds_number(self, value):
        if value < 1000 or value > 999999999:
            raise serializers.ValidationError(_("Invalid rounds number"))
        return value

    def validate_default_password(self, value):
        """Check password complexity."""
        password_validation.validate_password(value)
        return value

    def validate(self, data):
        """Custom validation method

        Depending on 'ldap_auth_method' value, we check for different
        required parameters.
        """
        errors = {}
        if data["sms_password_recovery"]:
            provider = data.get("sms_provider")
            if provider:
                sms_settings = sms_backends.get_backend_settings(provider)
                if sms_settings:
                    for name in sms_settings.keys():
                        if not data.get(name):
                            errors[name] = _("This field is required")
            else:
                errors["sms_provider"] = _("This field is required")

        if data["authentication_type"] == "ldap":
            if data["ldap_auth_method"] == "searchbind":
                required_fields = ["ldap_search_base", "ldap_search_filter"]
            else:
                required_fields = ["ldap_user_dn_template"]
            for f in required_fields:
                if data.get(f, "") == "":
                    errors[f] = _("This field is required")
        if len(errors):
            raise serializers.ValidationError(errors)
        return data


class LogSerializer(serializers.ModelSerializer):
    """Log serializer."""

    date_created = serializers.SerializerMethodField()

    class Meta:
        model = models.Log
        fields = ("date_created", "message", "level", "logger")

    def get_date_created(self, log) -> str:
        return formats.date_format(log.date_created, "SHORT_DATETIME_FORMAT")


class VerifyTFACodeSerializer(serializers.Serializer):
    """Serializer used to verify 2FA code validity."""

    code = serializers.CharField()

    def validate_code(self, value):
        device = django_otp.match_token(self.context["user"], value)
        if device is None:
            raise serializers.ValidationError(_("This code is invalid"))
        return device


class CheckPasswordSerializer(serializers.Serializer):
    """Simple serializer to check user password."""

    password = serializers.CharField()

    def validate_password(self, value):
        user = self.context["user"]
        if not user.check_password(value):
            raise serializers.ValidationError(_("Invalid password"))
        return value



"""Tests for core application."""

from io import StringIO

import httmock
from dateutil.relativedelta import relativedelta

from django.core import mail
from django.core import management
from django.test import TestCase
from django.urls import reverse
from django.utils import timezone

from modoboa.lib.tests import ModoTestCase
from .. import factories, mocks, models


class AuthenticationTestCase(ModoTestCase):

    """Validate authentication scenarios."""

    @classmethod
    def setUpTestData(cls):  # NOQA:N802
        """Create test data."""
        super(AuthenticationTestCase, cls).setUpTestData()
        cls.account = factories.UserFactory(
            username="user@test.com", groups=("SimpleUsers",)
        )

    def test_authentication(self):
        """Validate simple case."""
        self.client.logout()
        data = {"username": "user@test.com", "password": "toto"}
        response = self.client.post(reverse("core:login"), data)
        self.assertEqual(response.status_code, 302)
        self.assertTrue(response.url.endswith(reverse("core:user_index")))

        response = self.client.post(reverse("core:logout"), {})
        self.assertEqual(response.status_code, 302)

        data = {"username": "admin", "password": "password"}
        response = self.client.post(reverse("core:login"), data)
        self.assertEqual(response.status_code, 302)
        self.assertTrue(response.url.endswith(reverse("core:dashboard")))


class ManagementCommandsTestCase(TestCase):
    """Test management commands."""

    def test_change_default_admin(self):
        """Use dedicated option."""
        management.call_command(
            "load_initial_data", "--admin-username", "modoadmin")
        self.assertTrue(
            self.client.login(username="modoadmin", password="password"))

    def test_clean_logs(self):
        """Run cleanlogs command."""
        log1 = factories.LogFactory()
        factories.LogFactory()
        log1.date_created -= relativedelta(years=2)
        log1.save(update_fields=["date_created"])
        management.call_command("cleanlogs")
        self.assertEqual(models.Log.objects.count(), 1)

    def test_clean_inactive_accounts(self):
        """Run clean_inactive_accounts command."""
        management.call_command("load_initial_data")

        # no inactive account, should exit normaly
        management.call_command("clean_inactive_accounts")

        last_login = timezone.now() - relativedelta(days=45)
        account = factories.UserFactory(
            username="user1@domain.test", groups=("SimpleUsers", ),
            last_login=last_login
        )
        management.call_command("clean_inactive_accounts", "--dry-run")
        account.refresh_from_db()
        self.assertTrue(account.is_active)

        out = StringIO()
        management.call_command(
            "clean_inactive_accounts", "--verbose", "--dry-run", stdout=out)
        self.assertIn("user1@domain.test", out.getvalue())

        management.call_command("clean_inactive_accounts", "--silent")
        account.refresh_from_db()
        self.assertFalse(account.is_active)

        account.is_active = True
        account.save(update_fields=["is_active"])

        management.call_command(
            "clean_inactive_accounts", "--silent", "--delete")
        with self.assertRaises(models.User.DoesNotExist):
            account.refresh_from_db()


class ProfileTestCase(ModoTestCase):
    """Profile related tests."""

    @classmethod
    def setUpTestData(cls):  # NOQA:N802
        """Create test data."""
        super(ProfileTestCase, cls).setUpTestData()
        cls.account = factories.UserFactory(
            username="user@test.com", groups=("SimpleUsers",)
        )

    def test_update_profile(self):
        """Update profile without password."""
        data = {
            "first_name": "Homer", "last_name": "Simpson",
            "phone_number": "+33612345678", "language": "en"
        }
        self.ajax_post(reverse("core:user_profile"), data)
        admin = models.User.objects.get(username="admin")
        self.assertEqual(admin.last_name, "Simpson")

    def test_update_password(self):
        """Password update

        Two cases:
        * The default admin changes his password (no associated Mailbox)
        * A normal user changes his password
        """
        self.ajax_post(reverse("core:user_profile"),
                       {"language": "en", "oldpassword": "password",
                        "newpassword": "12345Toi", "confirmation": "12345Toi"})
        self.client.logout()

        self.assertEqual(
            self.client.login(username="admin", password="12345Toi"), True
        )
        self.assertEqual(
            self.client.login(username="user@test.com", password="toto"), True
        )

        self.ajax_post(
            reverse("core:user_profile"),
            {"oldpassword": "toto",
             "newpassword": "tutu", "confirmation": "tutu"},
            status=400
        )

        self.ajax_post(
            reverse("core:user_profile"),
            {"language": "en", "oldpassword": "toto",
             "newpassword": "Toto1234", "confirmation": "Toto1234"}
        )
        self.client.logout()
        self.assertTrue(
            self.client.login(username="user@test.com", password="Toto1234")
        )

    def test_update_password_url(self):
        """Check if external is used when defined."""
        self.set_global_parameter(
            "update_password_url", "http://update.password")
        non_local_user = factories.UserFactory(
            username="user@external.com", groups=("SimpleUsers",),
            is_local=False
        )
        self.client.force_login(non_local_user)
        url = reverse("core:user_profile")
        response = self.client.get(url)
        self.assertContains(response, "http://update.password")

        self.client.force_login(self.account)
        response = self.client.get(url)
        self.assertNotContains(response, "http://update.password")


class APIAccessFormTestCase(ModoTestCase):

    """Check form access."""

    @classmethod
    def setUpTestData(cls):  # NOQA:N802
        """Create test data."""
        super(APIAccessFormTestCase, cls).setUpTestData()
        cls.account = factories.UserFactory(
            username="user@test.com", groups=("SimpleUsers",)
        )

    def test_form_access(self):
        """Check access restrictions."""
        url = reverse("core:user_api_access")
        self.ajax_get(url)
        self.client.logout()
        self.client.login(username="user@test.com", password="toto")
        response = self.client.get(url, HTTP_X_REQUESTED_WITH="XMLHttpRequest")
        self.assertEqual(response.status_code, 278)

    def test_form(self):
        """Check that token is created/removed."""
        url = reverse("core:user_api_access")
        self.ajax_post(url, {"enable_api_access": True})
        user = models.User.objects.get(username="admin")
        self.assertTrue(hasattr(user, "auth_token"))
        self.ajax_post(url, {"enable_api_access": False})
        user = models.User.objects.get(username="admin")
        self.assertFalse(hasattr(user, "auth_token"))


class APICommunicationTestCase(ModoTestCase):
    """Check communication with the API."""

    def test_management_command(self):
        """Check command."""
        with httmock.HTTMock(
                mocks.modo_api_instance_search,
                mocks.modo_api_instance_create,
                mocks.modo_api_instance_update,
                mocks.modo_api_versions):
            management.call_command("communicate_with_public_api")
        self.assertEqual(models.LocalConfig.objects.first().api_pk, 100)
        self.assertEqual(len(mail.outbox), 0)

        url = reverse("core:information")
        response = self.ajax_request("get", url)
        self.assertIn("9.0.0", response["content"])

        # Enable notifications
        self.set_global_parameter("send_new_versions_email", True)
        with httmock.HTTMock(
                mocks.modo_api_instance_search,
                mocks.modo_api_instance_create,
                mocks.modo_api_instance_update,
                mocks.modo_api_versions):
            management.call_command("communicate_with_public_api")
        self.assertEqual(len(mail.outbox), 1)

        # Call once again and check no new notification has been sent
        self.set_global_parameter("send_new_versions_email", True)
        with httmock.HTTMock(
                mocks.modo_api_instance_search,
                mocks.modo_api_instance_create,
                mocks.modo_api_instance_update,
                mocks.modo_api_versions):
            management.call_command("communicate_with_public_api")
        self.assertEqual(len(mail.outbox), 1)

        # Make sure no new notification is sent when no updates
        with httmock.HTTMock(
                mocks.modo_api_instance_search,
                mocks.modo_api_instance_create,
                mocks.modo_api_instance_update,
                mocks.modo_api_versions_no_update):
            management.call_command("communicate_with_public_api")
        self.assertEqual(len(mail.outbox), 1)



"""Base core views."""

from django.contrib.auth import mixins as auth_mixins
from django.urls import reverse
from django.utils.http import is_safe_url
from django.views import generic

from ..extensions import exts_pool


def find_nextlocation(request, user):
    """Find next location for given user after login."""
    if not user.last_login:
        # Redirect to profile on first login
        return reverse("core:user_index")
    nextlocation = request.POST.get("next", request.GET.get("next"))
    condition = (
        nextlocation and
        is_safe_url(nextlocation, request.get_host())
    )
    if condition:
        return nextlocation
    if request.user.role == "SimpleUsers":
        topredir = request.localconfig.parameters.get_value(
            "default_top_redirection")
        if topredir != "user":
            infos = exts_pool.get_extension_infos(topredir)
            nextlocation = infos["topredirection_url"]
        else:
            nextlocation = reverse("core:user_index")
    else:
        nextlocation = reverse("core:dashboard")
    return nextlocation


class RootDispatchView(auth_mixins.LoginRequiredMixin, generic.RedirectView):
    """Handle root dispatching based on role."""

    def get_redirect_url(self):
        """Find proper next hop."""
        return find_nextlocation(self.request, self.request.user)



"""AppConfig for admin."""

from django.apps import AppConfig

from .app_settings import load_admin_settings


class AdminConfig(AppConfig):
    """App configuration."""

    default_auto_field = "django.db.models.BigAutoField"
    name = "modoboa.admin"
    verbose_name = "Modoboa admin console"

    def ready(self):
        load_admin_settings()

        from . import handlers  # NOQA:F401



"""Limits API."""

from django.contrib.contenttypes.models import ContentType

from rest_framework import mixins, viewsets
from rest_framework.permissions import DjangoModelPermissions, IsAuthenticated

from modoboa.core import models as core_models
from . import serializers


class ResourcesViewSet(
        mixins.RetrieveModelMixin,
        mixins.UpdateModelMixin,
        viewsets.GenericViewSet):
    """Resources viewset."""

    permission_classes = [IsAuthenticated, DjangoModelPermissions]
    serializer_class = serializers.ResourcesSerializer

    def get_queryset(self):
        """Filter queryset based on current user."""
        user = self.request.user
        ids = list(
            user.objectaccess_set.filter(
                content_type=ContentType.objects.get_for_model(user))
            .values_list("object_id", flat=True))
        if not user.is_superuser and user.pk in ids:
            ids.remove(user.pk)
        return core_models.User.objects.filter(pk__in=ids)



# Generated by Django 1.9.12 on 2017-02-06 06:13
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('admin', '0007_auto_20161104_1915'),
    ]

    operations = [
        migrations.AddField(
            model_name='domain',
            name='enable_dns_checks',
            field=models.BooleanField(default=True, help_text='Check to enable DNS checks for this domain', verbose_name='Enable DNS checks'),
        ),
    ]



"""Parameters management."""

import copy

from modoboa.lib import exceptions, form_utils, signals
from modoboa.lib.sysutils import guess_extension_name


class NotDefined(exceptions.ModoboaException):
    """Custom exception for undefined parameters."""

    http_code = 404

    def __init__(self, app, name=None):
        self.app = app
        self.name = name

    def __str__(self):
        if self.name is None:
            return "Application {} not registered".format(self.app)
        return "Parameter {} not defined for app {}".format(
            self.name, self.app)


class Registry(object):
    """A registry for parameters."""

    def __init__(self):
        """Constructor."""
        self._registry = {
            "global": {},
            "user": {}
        }
        self._registry2 = {
            "global": {},
            "user": {}
        }

    def add(self, level, formclass, label):
        """Add a new class containing parameters."""
        self._registry[level][formclass.app] = {
            "label": label, "formclass": formclass, "defaults": {}
        }

    def add2(self, level, app, label, structure, serializer_class):
        """Add new parameters for given app and level."""
        self._registry2[level][app] = {
            "label": label, "structure": structure,
            "serializer_class": serializer_class
        }

    def _load_default_values(self, level):
        """Load default values."""
        for _app, data in list(self._registry[level].items()):
            form = data["formclass"](load_values_from_db=False)
            for name, field in list(form.fields.items()):
                if isinstance(field, form_utils.SeparatorField):
                    continue
                data["defaults"][name] = field.initial

    def get_applications(self, level):
        """Return all applications registered for level."""
        result = [{"name": key, "label": value["label"]}
                  for key, value in self._registry2[level].items()]
        return result

    def get_forms(self, level, *args, **kwargs):
        """Return form instances for all app of the given level."""
        sorted_apps = []
        first_app = kwargs.pop("first_app", "core")
        if first_app in self._registry[level]:
            sorted_apps.append(first_app)
        sorted_apps += sorted(
            (
                a for a in self._registry[level]
                if a != first_app
            ),
            key=lambda a: self._registry[level][a]["label"]
        )
        result = []
        for app in sorted_apps:
            data = self._registry[level][app]
            if not data["formclass"].has_access(**kwargs):
                continue
            result.append({
                "app": app,
                "label": data["label"],
                "form": data["formclass"](*args, **kwargs)
            })
        return result

    def get_serializer_class(self, level, app):
        """Return serializer class for given level and app."""
        if app not in self._registry2[level]:
            raise NotDefined(app)
        return self._registry2[level][app]["serializer_class"]

    def get_structure(self, level, for_app=None):
        """Return fields structure."""
        result = []
        for app, fields in list(self._registry2[level].items()):
            if for_app and for_app != app:
                continue
            serializer = fields["serializer_class"]()
            for section, sconfig in list(fields["structure"].items()):
                item = {
                    "label": sconfig["label"],
                    "display": sconfig.get("display", ""),
                    "parameters": []
                }
                for name, config in list(sconfig["params"].items()):
                    data = {
                        "name": name,
                        "label": config["label"],
                        "help_text": config.get("help_text", ""),
                        "display": config.get("display", ""),
                    }
                    if config.get("separator"):
                        data["widget"] = "SeparatorField"
                    else:
                        field = serializer.fields[name]
                        if config.get("password"):
                            data["widget"] = "PasswordField"
                        else:
                            data["widget"] = field.__class__.__name__
                            if data["widget"] == "ChoiceField":
                                data["choices"] = [
                                    {"text": value, "value": key}
                                    for key, value in field.choices.items()
                                ]
                    item["parameters"].append(data)
                result.append(item)
        return result

    def exists(self, level, app, parameter=None):
        """Check if parameter exists."""
        parameters = self._registry[level]
        result = app in parameters
        if parameter:
            result = result and parameter in parameters[app]["defaults"]
        return result

    def get_default(self, level, app, parameter):
        """Retrieve default value for parameter."""
        if app not in self._registry[level]:
            raise NotDefined(app)
        if parameter not in self._registry[level][app]["defaults"]:
            raise NotDefined(app, parameter)
        return self._registry[level][app]["defaults"][parameter]

    def get_defaults(self, level, app):
        """Retrieve default values for application."""
        if app not in self._registry[level]:
            raise NotDefined(app)
        return self._registry[level][app]["defaults"]


registry = Registry()


class Manager:
    """A manager for parameters."""

    def __init__(self, level, parameters):
        """Constructor."""
        self._level = level
        self._parameters = parameters
        registry._load_default_values(level)

    def get_value(self, parameter, app=None, raise_exception=True):
        """Return the value associated to the specified parameter."""
        if app is None:
            app = guess_extension_name()
        # Compat. with the old way...
        parameter = parameter.lower()
        try:
            default = registry.get_default(self._level, app, parameter)
        except NotDefined:
            if raise_exception:
                raise
            return None
        if app in self._parameters:
            return self._parameters[app].get(parameter, default)
        return default

    def get_values(self, app=None):
        """Return all values for the given level/app."""
        if app is None:
            app = guess_extension_name()
        values = registry.get_defaults(self._level, app)
        for parameter, value in list(values.items()):
            if app in self._parameters:
                value = self._parameters[app].get(parameter, value)
            yield (parameter, value)

    def get_values_dict(self, app=None):
        """Return all values for the given app as a dictionary."""
        if app is None:
            app = guess_extension_name()
        values = registry.get_defaults(self._level, app)
        for parameter, value in list(values.items()):
            if app in self._parameters:
                values[parameter] = self._parameters[app].get(parameter, value)
        return values

    def set_value(self, parameter, value, app=None):
        """Set parameter for the given app."""
        if app is None:
            app = guess_extension_name()
        if not registry.exists(self._level, app, parameter):
            raise NotDefined(app, parameter)
        if app not in self._parameters:
            self._parameters[app] = copy.deepcopy(
                registry.get_defaults(self._level, app))
        self._parameters[app][parameter] = value

    def set_values(self, values, app=None):
        """Set/update values for the given app."""
        if app is None:
            app = guess_extension_name()
        if not registry.exists(self._level, app):
            raise NotDefined(app)
        if app not in self._parameters:
            self._parameters[app] = copy.deepcopy(
                registry.get_defaults(self._level, app))
        self._parameters[app].update(values)


def get_localconfig():
    """Retrieve current LocalConfig instance."""
    from modoboa.core import models as core_models
    request = signals.get_request()
    if request:
        return request.localconfig
    return core_models.LocalConfig.objects.first()


def get_global_parameter(name, app=None, **kwargs):
    """Retrieve a global parameter.

    This is a shortcut to use when a LocalConfig instance is not
    available (ie. when no request object is available).

    A ``NotDefined`` exception if the parameter doesn't exist.

    :param name: the parameter's name
    :param app: the application owning the parameter
    :return: the corresponding value
    """
    if app is None:
        app = guess_extension_name()
    return get_localconfig().parameters.get_value(name, app=app, **kwargs)


def get_global_parameters(app, **kwargs):
    """Retrieve global parameters of a given app.

    This is a shortcut to use when a LocalConfig instance is not
    available (ie. when no request object is available).

    A ``NotDefined`` exception if the parameter doesn't exist.

    :param name: the parameter's name
    :param app: the application owning the parameter
    :return: the corresponding value

    """
    return get_localconfig().parameters.get_values(app, **kwargs)


def apply_to_django_settings():
    """Apply global parameters to Django settings module."""
    for form in registry.get_forms("global"):
        form["form"].to_django_settings()



"""Tests for core application."""

import os
import smtplib
from unittest import mock, skipIf, skipUnless

from django.core import mail
from django.test import override_settings
from django.urls import reverse

try:
    import argon2
except ImportError:
    argon2 = None

from modoboa.core.password_hashers import (
    get_password_hasher, get_dovecot_schemes
)
from modoboa.lib.tests import NO_SMTP, ModoTestCase
from .. import factories, models

DOVEADM_TEST_PATH = "{}/doveadm".format(os.path.dirname(__file__))


class AuthenticationTestCase(ModoTestCase):
    """Validate authentication scenarios."""

    @classmethod
    def setUpTestData(cls):  # NOQA:N802
        """Create test data."""
        super(AuthenticationTestCase, cls).setUpTestData()
        cls.account = factories.UserFactory(
            username="user@test.com", groups=("SimpleUsers",)
        )

    def test_authentication(self):
        """Validate simple case."""
        self.client.logout()
        data = {"username": "user@test.com", "password": "toto"}
        response = self.client.post(reverse("core:login"), data)
        self.assertEqual(response.status_code, 302)
        self.assertTrue(response.url.endswith(reverse("core:user_index")))

        response = self.client.post(reverse("core:logout"), {})
        self.assertEqual(response.status_code, 302)

        data = {"username": "admin", "password": "password"}
        response = self.client.post(reverse("core:login"), data)
        self.assertEqual(response.status_code, 302)
        self.assertTrue(response.url.endswith(reverse("core:dashboard")))

    @mock.patch("django_otp.match_token")
    @mock.patch("django_otp.login")
    def test_authentication_with_2fa(self, login_mock, match_mock):
        user = models.User.objects.get(username="user@test.com")
        user.totpdevice_set.create(name="Device")
        user.tfa_enabled = True
        user.save()
        data = {"username": "user@test.com", "password": "toto"}
        response = self.client.post(reverse("core:login"), data)

        response = self.client.get(response.url)
        self.assertEqual(response.status_code, 302)
        url = reverse("core:2fa_verify")
        self.assertTrue(response.url.endswith(url))

        match_mock.side_effect = [None, user.totpdevice_set.first()]
        data = {"tfa_code": "1234"}
        response = self.client.post(url, data)
        self.assertContains(response, "This code is invalid")
        response = self.client.post(url, data)
        self.assertEqual(response.status_code, 302)

    def test_open_redirect(self):
        """Check that open redirect is not allowed."""
        self.client.logout()
        data = {"username": "admin", "password": "password"}

        # 1. Check valid redirection
        url = "{}?next=/admin/".format(reverse("core:login"))
        response = self.client.post(url, data)
        self.assertEqual(response.status_code, 302)
        self.assertTrue(response.url.endswith(reverse("admin:index")))
        self.client.logout()

        # 2. Check bad redirection
        url = "{}?next=http://www.evil.com".format(reverse("core:login"))
        response = self.client.post(url, data)
        self.assertEqual(response.status_code, 302)
        self.assertTrue(response.url.endswith(reverse("core:dashboard")))

    @override_settings(DOVEADM_LOOKUP_PATH=[DOVEADM_TEST_PATH])
    def test_password_schemes(self):
        """Validate password scheme changes."""
        username = "user@test.com"
        password = "toto"
        data = {"username": username, "password": password}
        user = models.User.objects.get(username=username)
        pw_hash = get_password_hasher('fallback_scheme')()

        self.client.logout()
        self.set_global_parameter("password_scheme", "sha512crypt")
        self.client.post(reverse("core:login"), data)
        user.refresh_from_db()
        self.assertTrue(user.password.startswith("{SHA512-CRYPT}"))

        self.client.logout()
        self.set_global_parameter("password_scheme", "sha256")
        self.client.post(reverse("core:login"), data)
        user.refresh_from_db()
        self.assertTrue(user.password.startswith("{SHA256}"))

        if argon2 is not None:
            self.client.logout()
            self.set_global_parameter("password_scheme", "argon2id")
            self.client.post(reverse("core:login"), data)
            user.refresh_from_db()
            self.assertTrue(user.password.startswith("{ARGON2ID}"))

        self.client.logout()
        self.set_global_parameter("password_scheme", "fallback_scheme")
        self.client.post(reverse("core:login"), data)
        user.refresh_from_db()
        self.assertTrue(user.password.startswith(pw_hash.scheme))

        self.client.logout()
        self.set_global_parameter("password_scheme", "sha256crypt")
        self.set_global_parameter("update_scheme", False)
        self.client.post(reverse("core:login"), data)
        user.refresh_from_db()
        self.assertTrue(user.password.startswith(pw_hash.scheme))

    @skipUnless(argon2, "argon2-cffi not installed")
    @override_settings(DOVEADM_LOOKUP_PATH=[DOVEADM_TEST_PATH])
    def test_password_argon2_parameter_change(self):
        """Validate hash parameter update on login works with argon2."""
        username = "user@test.com"
        password = "toto"
        data = {"username": username, "password": password}
        user = models.User.objects.get(username=username)
        self.set_global_parameter("password_scheme", "argon2id")

        self.client.logout()
        with self.settings(
            MODOBOA_ARGON2_TIME_COST=4,
            MODOBOA_ARGON2_MEMORY_COST=10000,
            MODOBOA_ARGON2_PARALLELISM=4,
        ):
            self.client.post(reverse("core:login"), data)
        user.refresh_from_db()
        self.assertTrue(user.password.startswith("{ARGON2ID}"))
        parameters = argon2.extract_parameters(
            user.password.lstrip("{ARGON2ID}")
        )
        self.assertEqual(parameters.time_cost, 4)
        self.assertEqual(parameters.memory_cost, 10000)
        self.assertEqual(parameters.parallelism, 4)

        self.client.logout()
        with self.settings(
            MODOBOA_ARGON2_TIME_COST=3,
            MODOBOA_ARGON2_MEMORY_COST=1000,
            MODOBOA_ARGON2_PARALLELISM=2,
        ):
            self.client.post(reverse("core:login"), data)
        user.refresh_from_db()
        self.assertTrue(user.password.startswith("{ARGON2ID}"))
        parameters = argon2.extract_parameters(
            user.password.lstrip("{ARGON2ID}"))
        self.assertEqual(parameters.time_cost, 3)
        self.assertEqual(parameters.memory_cost, 1000)
        self.assertEqual(parameters.parallelism, 2)

    @mock.patch("modoboa.lib.sysutils.exec_cmd")
    def test_dovecot_supported_schemes(self, exec_cmd_mock):
        """Validate Dovecot supported schemes with fake command output."""
        exec_cmd_mock.return_value = (
            0,
            "MD5 MD5-CRYPT SHA SHA1 SHA256 SHA512 SMD5 SSHA SSHA256 SSHA512 "
            "PLAIN CLEAR CLEARTEXT PLAIN-TRUNC CRAM-MD5 SCRAM-SHA-1 HMAC-MD5 "
            "DIGEST-MD5 PLAIN-MD4 PLAIN-MD5 LDAP-MD5 LANMAN NTLM OTP SKEY RPA "
            "PBKDF2 CRYPT SHA256-CRYPT SHA512-CRYPT"
        )
        supported_schemes = get_dovecot_schemes()
        self.assertEqual(supported_schemes,
                         ["{MD5}",
                          "{MD5-CRYPT}",
                          "{SHA}",
                          "{SHA1}",
                          "{SHA256}",
                          "{SHA512}",
                          "{SMD5}",
                          "{SSHA}",
                          "{SSHA256}",
                          "{SSHA512}",
                          "{PLAIN}",
                          "{CLEAR}",
                          "{CLEARTEXT}",
                          "{PLAIN-TRUNC}",
                          "{CRAM-MD5}",
                          "{SCRAM-SHA-1}",
                          "{HMAC-MD5}",
                          "{DIGEST-MD5}",
                          "{PLAIN-MD4}",
                          "{PLAIN-MD5}",
                          "{LDAP-MD5}",
                          "{LANMAN}",
                          "{NTLM}",
                          "{OTP}",
                          "{SKEY}",
                          "{RPA}",
                          "{PBKDF2}",
                          "{CRYPT}",
                          "{SHA256-CRYPT}",
                          "{SHA512-CRYPT}"])

    @override_settings(DOVECOT_SUPPORTED_SCHEMES="SHA1 SHA512-CRYPT")
    def test_dovecot_supported_schemes_from_settings(self):
        """Validate dovecot supported schemes from the settings."""
        supported_schemes = get_dovecot_schemes()
        self.assertEqual(supported_schemes, ["{SHA1}", "{SHA512-CRYPT}"])

    def test_dovecot_default_schemes(self):
        """Check default scheme if doveadm is not found."""
        supported_schemes = get_dovecot_schemes()
        self.assertEqual(supported_schemes, ["{MD5-CRYPT}", "{PLAIN}"])


class PasswordResetTestCase(ModoTestCase):
    """Test password reset service."""

    @classmethod
    def setUpTestData(cls):  # NOQA:N802
        """Create test data."""
        super(PasswordResetTestCase, cls).setUpTestData()
        cls.account_ok = factories.UserFactory(
            username="user@test.com", secondary_email="test@ext.com",
            phone_number="+33612345678",
            groups=("SimpleUsers",)
        )
        cls.account_ko = factories.UserFactory(
            username="user2@test.com", groups=("SimpleUsers",)
        )

    def test_reset_password(self):
        """Validate simple case."""
        self.client.logout()
        url = reverse("password_reset")
        data = {"email": self.account_ok.email}
        response = self.client.post(url, data, follow=True)
        self.assertContains(
            response,
            "We've emailed you instructions for setting your password")
        self.assertEqual(len(mail.outbox), 1)

    def test_reset_password_no_secondary_email(self):
        """Check email is not sent."""
        self.client.logout()
        url = reverse("password_reset")
        data = {"email": self.account_ko.email}
        response = self.client.post(url, data, follow=True)
        self.assertContains(
            response,
            "We've emailed you instructions for setting your password")
        self.assertEqual(len(mail.outbox), 0)

    def test_reset_password_announcement(self):
        """Check if announcement is showing."""
        self.client.logout()
        url = reverse("password_reset")
        msg = "Announcement"
        self.set_global_parameter("password_recovery_msg", msg)
        response = self.client.get(url)
        self.assertContains(response, msg)

    @mock.patch("oath.accept_totp")
    @mock.patch("ovh.Client.get")
    @mock.patch("ovh.Client.post")
    def test_reset_password_sms(self, client_post, client_get, accept_totp):
        """Test reset password by SMS."""
        client_get.return_value = ["service"]
        client_post.return_value = {"totalCreditsRemoved": 1}
        self.set_global_parameters({
            "sms_password_recovery": True,
            "sms_provider": "ovh",
            "sms_ovh_application_key": "key",
            "sms_ovh_application_secret": "secret",
            "sms_ovh_consumer_key": "consumer"
        })
        self.client.logout()
        url = reverse("password_reset")
        data = {"email": self.account_ok.email}
        response = self.client.post(url, data)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, reverse("password_reset_confirm_code"))

        data = {"code": "123456"}
        url = reverse("password_reset_confirm_code")
        accept_totp.return_value = (False, "")
        response = self.client.post(url, data)
        self.assertEqual(response.status_code, 200)
        accept_totp.return_value = (True, "")
        response = self.client.post(url, data)
        self.assertEqual(response.status_code, 302)

    @mock.patch("ovh.Client.get")
    @mock.patch("ovh.Client.post")
    def test_resend_reset_code(self, client_post, client_get):
        """Test resend code service."""
        url = reverse("password_reset_resend_code")
        # SMS password recovery not enabled
        response = self.client.get(url)
        self.assertEqual(response.status_code, 404)

        self.set_global_parameters({
            "sms_password_recovery": True,
            "sms_provider": "ovh",
            "sms_ovh_application_key": "key",
            "sms_ovh_application_secret": "secret",
            "sms_ovh_consumer_key": "consumer"
        }, app="core")
        # No user pk in session
        response = self.client.get(url)
        self.assertEqual(response.status_code, 404)

        session = self.client.session
        session["user_pk"] = self.account_ok.pk
        session.save()
        client_get.return_value = ["service"]
        client_post.return_value = {"totalCreditsRemoved": 1}
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)
        self.assertIn("totp_secret", self.client.session)


@skipIf(NO_SMTP, "No SMTP server available")
@override_settings(AUTHENTICATION_BACKENDS=(
    "modoboa.lib.authbackends.SMTPBackend",
    "django.contrib.auth.backends.ModelBackend"
))
class SMTPAuthenticationTestCase(ModoTestCase):
    """Validate SMTP authentication scenarios."""

    def _test_smtp_authentication(self, mock_smtp):
        """Common code to check authentication"""
        self.client.logout()
        username = "user@unknown.test"
        password = "toto"
        data = {"username": username, "password": password}
        response = self.client.post(reverse("core:login"), data)
        self.assertEqual(response.status_code, 302)
        self.assertTrue(response.url.endswith(reverse("core:user_index")))
        mock_smtp.return_value.login.assert_called_once_with(
            username, password)
        self.assertTrue(
            models.User.objects.filter(username=username).exists())

    @mock.patch("smtplib.SMTP")
    def test_smtp_authentication(self, mock_smtp):
        """Check simple SMTP authentication."""
        self._test_smtp_authentication(mock_smtp)

    @mock.patch("smtplib.SMTP_SSL")
    @override_settings(AUTH_SMTP_SECURED_MODE="ssl")
    def test_smtp_authentication_over_ssl(self, mock_smtp):
        """Check SMTP authentication over SSL."""
        self._test_smtp_authentication(mock_smtp)

    @mock.patch("smtplib.SMTP")
    @override_settings(AUTH_SMTP_SECURED_MODE="starttls")
    def test_smtp_authentication_over_starttls(self, mock_smtp):
        """Check SMTP authentication over STARTTLS."""
        self._test_smtp_authentication(mock_smtp)

    @mock.patch("smtplib.SMTP")
    def test_smtp_authentication_failure(self, mock_smtp):
        """Check SMTP authentication failure."""
        instance = mock_smtp.return_value
        instance.login.side_effect = smtplib.SMTPAuthenticationError(
            450, "User not found")
        self.client.logout()
        username = "user@unknown.test"
        password = "toto"
        data = {"username": username, "password": password}
        response = self.client.post(reverse("core:login"), data)
        self.assertEqual(response.status_code, 401)
        mock_smtp.return_value.login.assert_called_once_with(
            username, password)
        self.assertFalse(
            models.User.objects.filter(username=username).exists())



# Generated by Django 2.2.24 on 2021-11-08 15:52

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('maillog', '0002_auto_20200916_1346'),
    ]

    operations = [
        migrations.AlterField(
            model_name='maillog',
            name='date',
            field=models.DateTimeField(db_index=True),
        ),
    ]



"""Core API v2 views."""

import logging

from django.utils.html import escape
from django.utils.translation import ugettext as _

from rest_framework import response, status
from rest_framework_simplejwt import views as jwt_views
from rest_framework_simplejwt.exceptions import InvalidToken, TokenError

from modoboa.core.password_hashers import get_password_hasher
from modoboa.parameters import tools as param_tools

logger = logging.getLogger("modoboa.auth")


class TokenObtainPairView(jwt_views.TokenObtainPairView):
    """We overwrite this view to deal with password scheme update."""

    def post(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        try:
            serializer.is_valid(raise_exception=True)
        except TokenError as e:
            logger.warning(
                _("Failed connection attempt from '%s' as user '%s'"),
                request.META["REMOTE_ADDR"],
                escape(serializer.data["username"])
            )
            raise InvalidToken(e.args[0])

        user = serializer.user
        logger.info(
            _("User '%s' successfully logged in"), user.username
        )
        if user and user.is_active:
            condition = (
                user.is_local and
                param_tools.get_global_parameter(
                    "update_scheme", raise_exception=False)
            )
            if condition:
                # check if password scheme is correct
                scheme = param_tools.get_global_parameter(
                    "password_scheme", raise_exception=False)
                # use SHA512CRYPT as default fallback
                if scheme is None:
                    pwhash = get_password_hasher("sha512crypt")()
                else:
                    pwhash = get_password_hasher(scheme)()
                if not user.password.startswith(pwhash.scheme):
                    logger.info(
                        _("Password scheme mismatch. Updating %s password"),
                        user.username
                    )
                    user.set_password(request.data["password"])
                    user.save()
                if pwhash.needs_rehash(user.password):
                    logger.info(
                        _("Password hash parameter missmatch. "
                          "Updating %s password"),
                        user.username
                    )
                    user.set_password(serializer.data["password"])
                    user.save()

        return response.Response(
            serializer.validated_data, status=status.HTTP_200_OK)



"""Custom authentication backends."""

import smtplib

from django.conf import settings

from modoboa.core.models import User, populate_callback
from modoboa.lib.email_utils import split_mailbox
from modoboa.parameters import tools as param_tools


class SMTPBackend(object):
    """A backend to authenticate against an SMTP server."""

    def authenticate(self, request, username=None, password=None):
        """Check the username/password and return a User."""
        host = getattr(settings, "AUTH_SMTP_SERVER_ADDRESS", "localhost")
        port = getattr(settings, "AUTH_SMTP_SERVER_PORT", 25)
        secured_mode = getattr(settings, "AUTH_SMTP_SECURED_MODE", None)
        if secured_mode == "ssl":
            smtp = smtplib.SMTP_SSL(host, port)
        else:
            smtp = smtplib.SMTP(host, port)
            if secured_mode == "starttls":
                smtp.starttls()
        try:
            smtp.login(username, password)
        except smtplib.SMTPException:
            return None
        return self.get_or_build_user(username)

    def get_or_build_user(self, username):
        """Get a user or create it the first time.

        .. note::

           We assume the username is a valid email address.
        """
        user, created = User.objects.get_or_create(
            username__iexact=username, defaults={
                "username": username.lower(),
                "email": username.lower(),
                "language": settings.LANGUAGE_CODE
            }
        )
        if created:
            populate_callback(user)
        return user

    def get_user(self, user_pk):
        """Retrieve a User instance."""
        user = None
        try:
            user = User.objects.get(pk=user_pk)
        except User.DoesNotExist:
            pass
        return user


try:
    from django_auth_ldap.backend import (
        LDAPBackend as orig_LDAPBackend, _LDAPUser
    )

    class LDAPBackendBase(orig_LDAPBackend):

        def __init__(self, *args, **kwargs):
            """Load LDAP settings."""
            param_tools.apply_to_django_settings()
            super().__init__(*args, **kwargs)
            self.global_params = dict(
                param_tools.get_global_parameters("core"))

        def get_or_build_user(self, username, ldap_user):
            """
            This must return a (User, created) 2-tuple for the given
            LDAP user.  username is the Django-friendly username of
            the user. ldap_user.dn is the user's DN and
            ldap_user.attrs contains all of their LDAP attributes.
            """
            group = "SimpleUsers"
            admin_groups = self.global_params["ldap_admin_groups"].split(";")
            for grp in admin_groups:
                if grp.strip() in ldap_user.group_names:
                    group = "DomainAdmins"
                    break
            lpart, domain = split_mailbox(username)
            if domain is None:
                # Try to find associated email
                email = None
                for attr in ['mail', 'userPrincipalName']:
                    if attr in ldap_user.attrs:
                        email = ldap_user.attrs[attr][0]
                        break
                if email is None:
                    if group == "SimpleUsers":
                        # Only DomainAdmins can have a username which
                        # is not an email address
                        return None
                else:
                    username = email
            user, created = User.objects.get_or_create(
                username__iexact=username,
                defaults={
                    "username": username.lower(),
                    "is_local": False,
                    "language": settings.LANGUAGE_CODE
                }
            )
            if created:
                populate_callback(user, group)
            return user, created

        def get_user(self, user_id):
            user = None
            try:
                user = User.objects.get(pk=user_id)
                _LDAPUser(self, user=user)  # This sets user.ldap_user
            except User.DoesNotExist:
                pass
            return user

        def authenticate(self, *args, **kwargs):
            if self.global_params["authentication_type"] == "ldap":
                return super().authenticate(*args, **kwargs)
            return None

        @classmethod
        def setting_fullname(cls, setting):
            """Return fullname for given setting."""
            return "{}{}".format(cls.settings_prefix, setting)


    class LDAPBackend(LDAPBackendBase):
        """Primary LDAP backend."""

        settings_prefix = "AUTH_LDAP_"
        srv_address_setting_name = "ldap_server_address"
        srv_port_setting_name = "ldap_server_port"


    class LDAPSecondaryBackend(LDAPBackendBase):
        """Secondary LDAP backend."""

        settings_prefix = "AUTH_LDAP_2_"
        srv_address_setting_name = "ldap_secondary_server_address"
        srv_port_setting_name = "ldap_secondary_server_port"


except ImportError:
    pass



"""Custom context processors."""

from functools import reduce

from . import signals


def top_notifications(request):
    """A context processor to include top notifications."""
    if request.user.is_anonymous:
        return {}
    interval = request.localconfig.parameters.get_value(
        "top_notifications_check_interval")
    notifications = signals.get_top_notifications.send(
        sender="top_notifications", include_all=False)
    notifications = reduce(
        lambda a, b: a + b, [notif[1] for notif in notifications])
    return {
        "notifications_check_interval": interval * 1000,
        "top_notifications": notifications
    }



"""Utility functions."""

from pkg_resources import parse_version

from django.utils.translation import ugettext as _

from modoboa.core.extensions import exts_pool
from modoboa.lib.api_client import ModoAPIClient
from . import models


def parse_map_file(path):
    """Parse a postfix map file and return values."""
    content = {}
    with open(path) as fp:
        for line in fp:
            if not line or line.startswith("#"):
                continue
            name, value = line.split("=", 1)
            content[name.strip()] = value.strip()
    return content


def check_for_updates():
    """Check if a new version of Modoboa is available."""
    local_config = models.LocalConfig.objects.first()
    client = ModoAPIClient()
    extensions = exts_pool.list_all()
    extensions = [{
        "label": "Modoboa",
        "name": "modoboa",
        "description": _("The core part of Modoboa"),
        "version": client.local_core_version
    }] + extensions
    update_avail = False
    for extension in extensions:
        pkgname = extension["name"].replace("_", "-")
        for api_extension in local_config.api_versions:
            if api_extension["name"] != pkgname:
                continue
            extension["last_version"] = api_extension["version"]
            if (
                parse_version(api_extension["version"]) >
                parse_version(extension["version"])
            ):
                extension["update"] = True
                extension["changelog_url"] = api_extension["url"]
                update_avail = True
                break
    return update_avail, extensions



# Generated by Django 2.0.13 on 2020-01-04 17:21

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion

import modoboa.limits.models


class Migration(migrations.Migration):

    replaces = [
        ('limits', '0001_initial'),
        ('limits', '0002_auto_20151114_1518'),
        ('limits', '0003_auto_20160413_1046'),
        ('limits', '0004_auto_20160413_1312'),
    ]

    initial = True

    dependencies = [
        ('admin', '0001_initial'),
        ('contenttypes', '0002_remove_content_type_name'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='UserObjectLimit',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=254)),
                ('max_value', models.IntegerField(default=0)),
                ('content_type', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='contenttypes.ContentType')),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL)),
            ],
        ),
        migrations.AlterUniqueTogether(
            name='userobjectlimit',
            unique_together=set([('user', 'name')]),
        ),
        migrations.CreateModel(
            name='DomainObjectLimit',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('name', models.CharField(max_length=254)),
                ('max_value', models.IntegerField(default=0)),
                ('domain', models.ForeignKey(to='admin.Domain', on_delete=models.CASCADE)),
            ],
            bases=(modoboa.limits.models.ObjectLimitMixin, models.Model),
        ),
        migrations.AlterUniqueTogether(
            name='domainobjectlimit',
            unique_together=set([('domain', 'name')]),
        ),
    ]



# Generated by Django 1.9.5 on 2016-10-26 08:03
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('admin', '0004_auto_20160914_0907'),
    ]

    operations = [
        migrations.AlterField(
            model_name='alias',
            name='description',
            field=models.TextField(blank=True, verbose_name='Description'),
        ),
        migrations.AlterField(
            model_name='alias',
            name='expire_at',
            field=models.DateTimeField(blank=True, null=True, verbose_name='Expire at'),
        ),
    ]



"""Tools for top level testing (ie. when ran by Travis)."""

import os
import shutil
import tempfile

from django.core.management import call_command


class MapFilesTestCaseMixin(object):
    """A generic test case to check map files generation."""

    MAP_FILES = None

    extension = None

    def setUp(self):
        self.workdir = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.workdir)

    def _test_maps_generation(self, engine):
        dburl = "{0}://user:password@localhost/testdb".format(engine)
        call_command(
            "generate_postfix_maps",
            "--dburl", dburl, "--destdir", self.workdir)

        for mapfile in self.MAP_FILES:
            path = "{0}/{1}".format(self.workdir, mapfile)
            self.assertTrue(os.path.exists(path))
            with open(path) as fpo:
                content = fpo.read()
            if engine != "sqlite":
                self.assertIn("user = user", content)
                self.assertIn("password = password", content)
                self.assertIn("dbname = testdb", content)
                self.assertIn("hosts = localhost", content)
            else:
                self.assertIn("dbpath = testdb", content)

    def test_postgres_maps(self):
        self._test_maps_generation("postgres")

    def test_mysql_maps(self):
        self._test_maps_generation("mysql")

    def test_sqlite_maps(self):
        self._test_maps_generation("sqlite")



"""
Custom pagination tools.
"""

class Page(object):

    """Class to represent a page."""

    def __init__(self, paginator, pageid, id_start, id_stop, baseurl=None):
        """Constructor."""
        self.paginator = paginator
        self.number = pageid
        self.id_start = id_start
        self.id_stop = id_stop
        self.baseurl = baseurl

        self._has_previous = None
        self._has_next = None

    @property
    def items_per_page(self):
        """The number of items in a page. (shortcut)"""
        return self.paginator.elems_per_page

    @property
    def items(self):
        """The number of items in this page."""
        return self.id_stop - self.id_start

    @property
    def has_previous(self):
        """Tell if a previous page is available or not."""
        if self._has_previous is None:
            self._has_previous = True if self.number > 1 else False
        return self._has_previous

    @property
    def previous_page_number(self):
        """Return the index of the previous page."""
        if not self.has_previous:
            return False
        return self.number - 1

    @property
    def has_next(self):
        """Tell if a next page is available or not."""
        if self._has_next is None:
            self._has_next = (
                True if self.id_stop < self.paginator.total else False
            )
        return self._has_next

    @property
    def next_page_number(self):
        """Return the index of the next page."""
        if not self.has_next:
            return False
        return self.number + 1

    @property
    def last_page(self):
        """Retrieve the id of the last page."""
        lid = self.paginator.total / self.items_per_page
        if not lid:
            return 1
        if self.paginator.total % self.items_per_page:
            lid += 1
        return lid


class Paginator(object):

    """Pagination class."""

    def __init__(self, total, elems_per_page):
        self.total = total
        self.elems_per_page = elems_per_page
        self.num_pages = total / elems_per_page
        if total % elems_per_page:
            self.num_pages += 1

    def _indexes(self, page):
        """Compute start and stop indexes."""
        id_start = self.elems_per_page * page + 1
        id_stop = id_start + self.elems_per_page - 1
        return (id_start, id_stop)

    def getpage(self, page_id):
        """Retrieve a specific page."""
        if page_id < 1:
            return None
        id_start, id_stop = self._indexes(page_id - 1)
        if id_start <= self.total:
            if id_stop >= self.total:
                id_stop = self.total
        else:
            return None
        return Page(self, page_id, id_start, id_stop)



# Generated by Django 1.9.5 on 2016-11-04 21:06
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('lib', '0005_auto_20160416_1449'),
    ]

    operations = [
        migrations.DeleteModel(
            name='Parameter',
        ),
        migrations.RemoveField(
            model_name='userparameter',
            name='user',
        ),
        migrations.DeleteModel(
            name='UserParameter',
        ),
    ]



"""Parsing tools for maillog related files."""

import io
import re
import time

from modoboa.admin import models as admin_models
from modoboa.lib.email_utils import split_mailbox

from . import utils


class MaillogParser:
    """Main parser class for maillog file."""

    def __init__(self, year=None, greylist=False, verbose=False, debug=False):
        """Constructor."""
        self.debug = debug
        self.verbose = verbose

        self.domains = []
        self._load_domain_list()

        self.workdict = {}
        self.lupdates = {}

        self.__year = year
        curtime = time.localtime()
        if not self.__year:
            self.__year = curtime.tm_year
        self.curmonth = curtime.tm_mon

        # set up regular expression
        self._date_expressions = [
            r"(?P<month>\w+)\s+(?P<day>\d+)\s+(?P<hour>\d+):(?P<min>\d+):(?P<sec>\d+)(?P<eol>.*)",  # noqa
            r"(?P<year>\d+)-(?P<month>\d+)-(?P<day>\d+)T(?P<hour>\d+):(?P<min>\d+):(?P<sec>\d+)\.\d+\+\d+:\d+(?P<eol>.*)"  # noqa
        ]
        self._date_expressions = [
            re.compile(v) for v in self._date_expressions]
        self.date_expr = None
        self._regex = {
            "line": r"\s+([-\w\.]+)\s+(\w+)/?(\w*)[[](\d+)[]]:\s+(.*)",
            "id": r"(\w+): (.*)",
            "reject": r"reject: .*from=<.*>,? to=<[^@]+@([^>]+)>",
            "message-id": r"message-id=<([^>]*)>",
            "from+size": r"from=<([^>]*)>, size=(\d+)",
            "to+status": r"to=<([^>]*)>.*status=(\S+)",
            "orig_to": r"orig_to=<([^>]*)>.*",
        }
        self._regex = {k: re.compile(v) for k, v in self._regex.items()}
        self._srs_regex = {
            "detect_srs": "^SRS[01][+-=]",
            "reverse_srs0": r"^SRS0[+-=]\S+=\S{2}=(\S+)=(\S+)\@\S+$",
            "reverse_srs1": r"^SRS1[+-=]\S+=\S+==\S+=\S{2}=(\S+)=(\S+)\@\S+$"
        }
        self._srs_regex = {
            k: re.compile(v, re.IGNORECASE) for k, v in self._srs_regex.items()
        }

        self.greylist = greylist
        if greylist:
            self._dprint("[settings] greylisting enabled")

    def _load_domain_list(self):
        """Load the list of allowed domains."""
        for dom in admin_models.Domain.objects.all():
            domname = str(dom.name)
            self.domains += [domname]
            # Also add alias domains
            for alias in dom.domainalias_set.all():
                aliasname = str(alias.name)
                self.domains += [aliasname]

    def _dprint(self, msg):
        """Print a debug message if required.

        :param str msg: debug message
        """
        if not self.debug:
            return
        print(msg)

    def year(self, month):
        """Return the appropriate year

        Date used in log files don't always embark the year so we need
        to guess it :p

        This method tries to deal with year changes in a simply
        (ugly?) way: if we currently are in january and the given
        month is different, return the current year -1. Otherwise,
        return the current year.

        Obviously, this method only works for year to year + 1
        changes.

        :param month: the month of the current record beeing parsed
        :return: an integer
        """
        month = (
            time.strptime(month, "%b").tm_mon if not month.isdigit()
            else int(month)
        )
        if self.curmonth == 1 and month != self.curmonth:
            return self.__year - 1
        return self.__year

    def _store_current_date(self, match):
        """Transform and store parsed date."""
        ho = match.group("hour")
        mi = match.group("min")
        se = match.group("sec")
        mo = match.group("month")
        da = match.group("day")
        try:
            ye = match.group("year")
        except IndexError:
            ye = self.year(mo)
        self.cur_t = utils.date_to_timestamp([ye, mo, da, ho, mi, se])

    def _parse_date(self, line):
        """Try to match a date inside :kw:`line` and to convert it to
        a timestamp.

        We try different date format until we find valid one. We then
        store it for future use.

        :param str line: a log entry
        :return: the remaining part of the line or None
        """
        match = None
        if self.date_expr is None:
            for expr in self._date_expressions:
                match = expr.match(line)
                if match is not None:
                    self.date_expr = expr
                    break
        else:
            match = self.date_expr.match(line)
        if match is None:
            return None
        self._store_current_date(match)
        return match.group("eol")

    def is_srs_forward(self, mail_address):
        """ Check if mail address has been mangled by SRS

        Sender Rewriting Scheme (SRS) modifies mail adresses so that they
        always end in a local domain.

        :param str mail_address
        :return a boolean
        """
        return self._srs_regex["detect_srs"].match(mail_address) is not None

    def reverse_srs(self, mail_address):
        """ Try to unwind a mail address rewritten by SRS

        Sender Rewriting Scheme (SRS) modifies mail adresses so that they
        always end in a local domain. Common Postfix implementations of SRS
        rewrite all non-local mail addresss.

        :param str mail_address
        :return a str
        """
        m = self._srs_regex["reverse_srs0"].match(mail_address)
        m = self._srs_regex["reverse_srs1"].match(mail_address) \
            if m is None else m

        if m is not None:
            return "%s@%s" % m.group(2, 1)
        return mail_address

    def new_domain_event(self, domain, name, size=None):
        """Take action about new event for domain."""
        pass

    def new_message_processed(
            self, queue_id, msg_status, from_domain, to_domain, msg_to,
            msg_orig_to):
        """Store new message in local database."""
        pass

    def _parse_postfix(self, log, host, pid, subprog):
        """ Parse a log entry generated by Postfix.

        :param str log: logged message
        :param str host: hostname
        :param str pid: process ID
        :return: True on success
        """
        m = self._regex["id"].match(log)
        if m is None:
            return False

        queue_id, msg = m.groups()

        # Handle rejected mails.
        if queue_id == "NOQUEUE":
            m = self._regex["reject"].match(msg)
            dom = m.group(1) if m is not None else None
            if dom in self.domains:
                condition = (
                    self.greylist and (
                        "Greylisted" in msg or
                        (subprog == "postscreen" and " 450 " in msg))
                )
                self.new_domain_event(
                    dom, "greylist" if condition else "reject")
            return True

        # Message acknowledged.
        m = self._regex["message-id"].search(msg)
        if m is not None:
            self.workdict[queue_id] = {"from": m.group(1), "size": 0}
            return True

        # Message enqueued.
        m = self._regex["from+size"].search(msg)
        if m is not None:
            self.workdict[queue_id] = {
                "from": self.reverse_srs(m.group(1)),
                "size": int(m.group(2))
            }
            return True

        # Message disposition.
        m = self._regex["to+status"].search(msg)
        if m is None:
            return False
        (msg_to, msg_status) = m.groups()
        if queue_id not in self.workdict:
            self._dprint("[parser] inconsistent mail (%s: %s), skipping"
                         % (queue_id, msg_to))
            return True

        # orig_to is optional.
        m = self._regex["orig_to"].search(msg)
        msg_orig_to = m.group(1) if m is not None else None

        # Handle local "from" domains.
        from_domain = split_mailbox(self.workdict[queue_id]["from"])[1]
        if from_domain is not None and from_domain in self.domains:
            self.new_domain_event(
                from_domain, "sent", self.workdict[queue_id]["size"])

        # Handle local "to" domains.
        to_domain = None
        condition = (
            msg_orig_to is not None and
            not self.is_srs_forward(msg_orig_to)
        )
        if condition:
            to_domain = split_mailbox(msg_orig_to)[1]
        if to_domain is None:
            to_domain = split_mailbox(msg_to)[1]

        if msg_status == "sent":
            msg_status = "recv"
        self.new_domain_event(
            to_domain, "recv", self.workdict[queue_id]["size"])

        # Store log entry
        self.new_message_processed(
            queue_id, msg_status, from_domain, to_domain, msg_to, msg_orig_to)

        return True

    def _parse_line(self, line):
        """Parse a single log line.

        :param str line: log line
        """
        line = self._parse_date(line)
        if line is None:
            return
        m = self._regex["line"].match(line)
        if not m:
            return
        host, prog, subprog, pid, log = m.groups()

        try:
            parser = getattr(self, "_parse_{}".format(prog))
            if not parser(log, host, pid, subprog):
                self._dprint("[parser] ignoring %r log: %r" % (prog, log))
        except AttributeError:
            self._dprint(
                "[parser] no log handler for \"%r\": %r".format(prog, log))

    def parse(self, logfile):
        """Process the log file."""
        try:
            fp = io.open(logfile, encoding="utf-8")
        except IOError as errno:
            self._dprint("%s" % errno)
            return

        for line in fp.readlines():
            self._parse_line(line)

        fp.close()



# Generated by Django 2.1.15 on 2020-01-13 09:22

import sys

from django.db import IntegrityError, migrations, transaction

VIEW_PERMISSION_CODENAMES = {
    'view_aliases': 'view_alias',
    'view_domaliases': 'view_domainalias',
    'view_mailboxes': 'view_mailbox',
}


def rename_view_permissions(apps, schema_editor):
    Permission = apps.get_model('auth', 'Permission')
    db_alias = schema_editor.connection.alias
    # Iterate over permissions and try to rename the codename
    for perm in Permission.objects.using(db_alias).filter(
        codename__in=VIEW_PERMISSION_CODENAMES.keys()
    ):
        codename = perm.codename
        try:
            with transaction.atomic():
                perm.codename = VIEW_PERMISSION_CODENAMES[perm.codename]
                perm.save()
        except IntegrityError:
            print(
                (
                    "Unable to rename '{}' to '{}', a permission with this "
                    "codename already exists.".format(codename, perm.codename)
                ),
                file=sys.stderr,
            )


class Migration(migrations.Migration):

    dependencies = [
        ('admin', '0014_auto_20181017_1628'),
    ]

    operations = [
        migrations.RunPython(rename_view_permissions),
    ]



"""App related urls."""

from rest_framework import routers

from . import viewsets


router = routers.SimpleRouter()
router.register(
    r"transports", viewsets.TransportViewSet, basename="transport")

urlpatterns = router.urls



"""Management command to communicate with public API."""

from django.core.mail import EmailMessage
from django.core.management.base import BaseCommand, CommandError
from django.template.loader import render_to_string
from django.utils.translation import ugettext as _

from modoboa.admin import models as admin_models
from modoboa.core.extensions import exts_pool
from modoboa.lib import api_client
from ... import models
from ... import utils


class Command(BaseCommand):
    """Command class."""

    help = "Communicate with Modoboa public API."  # NOQA:A003

    def send_notification(self, local_config, extensions):
        """Send email notification about updates."""
        updates = {}
        to_create = []
        for extension in extensions:
            if "update" not in extension:
                continue
            qset = models.ExtensionUpdateHistory.objects.filter(
                extension=extension["name"], version=extension["last_version"]
            )
            if not qset.exists():
                updates[extension["name"]] = extension["last_version"]
                to_create.append(
                    models.ExtensionUpdateHistory(
                        extension=extension["name"],
                        version=extension["last_version"]
                    )
                )
        if not updates:
            return
        content = render_to_string(
            "core/notifications/update_available.html", {
                "updates": updates
            })
        subject = _("[modoboa] Update(s) available")
        sender = local_config.parameters.get_value("sender_address")
        recipient = local_config.parameters.get_value(
            "new_versions_email_rcpt")
        msg = EmailMessage(
            subject, content.strip(), sender, [recipient],
        )
        msg.send()
        models.ExtensionUpdateHistory.objects.bulk_create(to_create)

    def handle(self, *args, **options):
        """Command entry point."""
        local_config = (
            models.LocalConfig.objects.select_related("site").first())
        if not local_config.parameters.get_value("enable_api_communication"):
            return
        self.client = api_client.ModoAPIClient()
        if not local_config.api_pk:
            pk = self.client.register_instance(local_config.site.domain)
            if pk is None:
                raise CommandError("Instance registration failed.")
            local_config.api_pk = pk

        if local_config.parameters.get_value("check_new_versions"):
            versions = self.client.versions()
            if versions is None:
                raise CommandError("Failed to retrieve versions from the API.")
            local_config.api_versions = versions

        local_config.save()

        if local_config.parameters.get_value("send_new_versions_email"):
            update_avail, extensions = utils.check_for_updates()
            if update_avail:
                self.send_notification(local_config, extensions)

        if not local_config.parameters.get_value("send_statistics"):
            return
        extensions = [ext["name"] for ext in exts_pool.list_all()]
        data = {
            "hostname": local_config.site.domain,
            "known_version": self.client.local_core_version,
            "domain_counter": admin_models.Domain.objects.count(),
            "domain_alias_counter": admin_models.DomainAlias.objects.count(),
            "mailbox_counter": admin_models.Mailbox.objects.count(),
            "alias_counter": (
                admin_models.Alias.objects.filter(internal=False).count()),
            "extensions": extensions
        }
        self.client.update_instance(local_config.api_pk, data)



"""DNSBL related views."""

from django.contrib.auth import mixins as auth_mixins
from django.utils.translation import ugettext as _
from django.views import generic

from .. import models


class MXDomainDetailView(
        auth_mixins.PermissionRequiredMixin, generic.DetailView):
    """View to display MX records."""

    model = models.Domain
    permission_required = "admin.view_domain"
    template_name = "admin/mx_domain_detail.html"

    def get_queryset(self):
        """Add some prefetching."""
        return (
            super(MXDomainDetailView, self).get_queryset()
            .prefetch_related("mxrecord_set"))

    def get_context_data(self, **kwargs):
        """Add extra variables."""
        context = super(MXDomainDetailView, self).get_context_data(**kwargs)
        context.update({
            "title": _("MX records of {}").format(self.object.name)
        })
        return context


class DNSBLDomainDetailView(
        auth_mixins.PermissionRequiredMixin, generic.DetailView):
    """View to display DNSBL summary."""

    model = models.Domain
    permission_required = "admin.view_domain"
    template_name = "admin/dnsbl_domain_detail.html"

    def get_queryset(self):
        """Add some prefetching."""
        return (
            super(DNSBLDomainDetailView, self).get_queryset()
            .prefetch_related("dnsblresult_set"))

    def get_context_data(self, **kwargs):
        """Add extra variables."""
        context = super(DNSBLDomainDetailView, self).get_context_data(**kwargs)
        context.update({
            "title": _("DNSBL summary for {}").format(self.object.name)
        })
        return context



# Generated by Django 2.0.13 on 2020-01-04 17:14

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0017_auto_20190705_1045'),
    ]

    operations = [
        migrations.AlterField(
            model_name='user',
            name='last_name',
            field=models.CharField(blank=True, max_length=150, verbose_name='last name'),
        ),
    ]



"""App related serializers."""

from rest_framework import serializers

from modoboa.admin import models as admin_models

from ... import models


class MXRecordSerializer(serializers.ModelSerializer):
    """Serializer for MXRecord."""

    class Meta:
        model = admin_models.MXRecord
        fields = ("name", "address", "updated")


class DNSBLResultSerializer(serializers.ModelSerializer):
    """Serializer for DNSBLResult."""

    mx = MXRecordSerializer()

    class Meta:
        model = admin_models.DNSBLResult
        fields = ("provider", "mx", "status")


class DNSRecordSerializer(serializers.ModelSerializer):
    """Serializer for DNSRecord."""

    class Meta:
        model = models.DNSRecord
        fields = ("type", "value", "is_valid", "error", "updated")


class DNSDetailSerializer(serializers.ModelSerializer):

    mx_records = MXRecordSerializer(many=True, source="mxrecord_set")
    autoconfig_record = DNSRecordSerializer()
    autodiscover_record = DNSRecordSerializer()
    spf_record = DNSRecordSerializer()
    dkim_record = DNSRecordSerializer()
    dmarc_record = DNSRecordSerializer()
    dnsbl_results = DNSBLResultSerializer(many=True, source="dnsblresult_set")

    class Meta:
        model = admin_models.Domain
        fields = (
            "mx_records", "dnsbl_results", "autoconfig_record",
            "autodiscover_record", "spf_record", "dkim_record", "dmarc_record"
        )



from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0004_auto_20150728_1228'),
    ]

    operations = [
        migrations.AddField(
            model_name='user',
            name='language',
            field=models.CharField(default=b'en', help_text='Prefered language to display pages.', max_length=10, choices=[(b'cs', '\u010de\u0161tina'), (b'de', 'deutsch'), (b'en', 'english'), (b'es', 'espa\xf1ol'), (b'fr', 'fran\xe7ais'), (b'it', 'italiano'), (b'nl', 'nederlands'), (b'pt_PT', 'portugu\xeas'), (b'pt_BR', 'portugu\xeas (BR)'), (b'ru', '\u0440\u0443\u0441\u0441\u043a\u0438\u0439'), (b'sv', 'svenska')]),
            preserve_default=True,
        ),
    ]



"""Policy daemon management command."""

import asyncio
from contextlib import suppress
import functools
import logging
import signal

from django.core.management.base import BaseCommand

from ... import core

logger = logging.getLogger("modoboa.policyd")


def ask_exit(signame, loop):
    """Stop event loop."""
    loop.stop()


class Command(BaseCommand):
    """Management command for policy daemon."""

    help = "Launch Modoboa policy daemon"

    def add_arguments(self, parser):
        """Add command line arguments."""
        parser.add_argument("--host", type=str, default="localhost")
        parser.add_argument("--port", type=int, default=9999)
        parser.add_argument("--debug", action="store_true",
                            help="Enable debug mode")

    def handle(self, *args, **options):
        """Entry point."""
        loop = asyncio.get_event_loop()
        coro = asyncio.start_server(
            core.new_connection, options["host"], options["port"], loop=loop
        )
        server = loop.run_until_complete(coro)

        # Schedule reset task
        core.start_reset_counters_coro()

        for signame in {'SIGINT', 'SIGTERM'}:
            loop.add_signal_handler(
                getattr(signal, signame),
                functools.partial(ask_exit, signame, loop)
            )

        logger.info("Serving on {}".format(server.sockets[0].getsockname()))

        if options["debug"]:
            loop.set_debug(True)
            logging.getLogger('asyncio').setLevel(logging.DEBUG)

        loop.run_forever()

        logger.info("Stopping policy daemon...")
        # Close the server
        server.close()
        loop.run_until_complete(server.wait_closed())
        # Cancel pending tasks
        for task in asyncio.all_tasks(loop):
            task.cancel()
            # Await task to execute it's cancellation. Cancelled task
            # raises asyncio.CancelledError that we can suppress
            with suppress(asyncio.CancelledError):
                loop.run_until_complete(task)
        loop.close()



"""Core dashboard views."""

import feedparser
import requests
from dateutil import parser
from requests.exceptions import RequestException

from django.contrib.auth import mixins as auth_mixins
from django.views import generic

from django.conf import settings

from .. import signals

MODOBOA_WEBSITE_URL = "https://modoboa.org/"


class DashboardView(auth_mixins.AccessMixin, generic.TemplateView):
    """Dashboard view."""

    template_name = "core/dashboard.html"

    def dispatch(self, request, *args, **kwargs):
        """Check if user can access dashboard."""
        if not request.user.is_authenticated or not request.user.is_admin:
            return self.handle_no_permission()
        return super(DashboardView, self).dispatch(request, *args, **kwargs)

    def get_context_data(self, **kwargs):
        """Add context variables."""
        context = super(DashboardView, self).get_context_data(**kwargs)
        context.update({
            "selection": "dashboard", "widgets": {"left": [], "right": []}
        })
        # Fetch latest news
        if self.request.user.language == "fr":
            lang = "fr"
        else:
            lang = "en"
        context.update({"selection": "dashboard"})

        feed_url = "{}{}/weblog/feeds/".format(MODOBOA_WEBSITE_URL, lang)
        if self.request.user.role != "SuperAdmins":
            custom_feed_url = (
                self.request.localconfig.parameters.get_value("rss_feed_url"))
            if custom_feed_url:
                feed_url = custom_feed_url
        entries = []
        if not settings.DISABLE_DASHBOARD_EXTERNAL_QUERIES:
            posts = feedparser.parse(feed_url)
            for entry in posts["entries"][:5]:
                entry["published"] = parser.parse(entry["published"])
                entries.append(entry)
        context["widgets"]["left"].append("core/_latest_news_widget.html")
        context.update({"news": entries})

        hide_features_widget = self.request.localconfig.parameters.get_value(
            "hide_features_widget")
        if self.request.user.is_superuser or not hide_features_widget:
            url = "{}{}/api/projects/?featured=true".format(
                MODOBOA_WEBSITE_URL, lang)
            features = []
            if not settings.DISABLE_DASHBOARD_EXTERNAL_QUERIES:
                try:
                    response = requests.get(url)
                except RequestException:
                    pass
                else:
                    if response.status_code == 200:
                        features = response.json()
            context["widgets"]["right"].append("core/_current_features.html")
            context.update({"features": features})

        # Extra widgets
        result = signals.extra_admin_dashboard_widgets.send(
            sender=self.__class__, user=self.request.user)
        for _receiver, widgets in result:
            for widget in widgets:
                context["widgets"][widget["column"]].append(
                    widget["template"])
                # FIXME: can raise conflicts...
                context.update(widget["context"])

        return context



# Generated by Django 1.10.7 on 2017-11-23 15:53
from django.db import migrations


def move_relaydomain_to_transport(apps, schema_editor):
    """Transform relaydomains to transports."""
    RelayDomain = apps.get_model("relaydomains", "RelayDomain")
    RecipientAccess = apps.get_model("relaydomains", "RecipientAccess")
    Transport = apps.get_model("transport", "Transport")
    ra_to_create = []
    for rd in RelayDomain.objects.select_related("domain", "service"):
        next_hop = "[{}]:{}".format(rd.target_host, rd.target_port)
        tr = Transport.objects.create(
            pattern=rd.domain.name,
            service="relay",
            next_hop=next_hop,
            _settings={
                "relay_target_host": rd.target_host,
                "relay_target_port": rd.target_port,
                "relay_verify_recipients": rd.verify_recipients
            }
        )
        rd.domain.transport = tr
        rd.domain.save(update_fields=["transport"])
        if not rd.verify_recipients:
            continue
        ra_to_create.append(
            RecipientAccess(
                pattern=rd.domain.name, action="reject_unverified_recipient"))
    RecipientAccess.objects.bulk_create(ra_to_create)


def forward(apps, schema_editor):
    """Empty."""
    pass


class Migration(migrations.Migration):

    dependencies = [
        ('relaydomains', '0007_recipientaccess'),
        ('transport', '0001_initial'),
        ('admin', '0011_domain_transport'),
    ]

    operations = [
        migrations.RunPython(move_relaydomain_to_transport, forward)
    ]



"""Transport backend."""

from django import forms
from django.core import validators
from django.utils.translation import ugettext_lazy as _

from modoboa.lib import validators as lib_validators


class TransportBackend(object):
    """Base backend class."""

    name = None
    settings = ()

    def _validate_host_address(self, value):
        """Validate an host address syntax."""
        validator_list = [
            lib_validators.validate_hostname,
            validators.validate_ipv46_address
        ]
        for validator in validator_list:
            try:
                validator(value)
            except forms.ValidationError:
                pass
            else:
                return True
        return False

    def clean_fields(self, values):
        """Check that values are correct."""
        errors = []
        for setting in self.settings:
            fname = "{}_{}".format(self.name, setting["name"])
            value = values.get(fname)
            if not value:
                if not setting.get("required", True):
                    continue
                errors.append((fname, _("This field is required")))
                continue
            ftype = setting.get("type", "string")
            if ftype == "string":
                validator = setting.get("validator")
                vfunc = "_validate_{}".format(validator)
                if validator and hasattr(self, vfunc):
                    if getattr(self, vfunc)(value):
                        continue
                else:
                    continue
            elif ftype == "int" and isinstance(value, int):
                continue
            elif ftype == "boolean" and isinstance(value, bool):
                continue
            errors.append((fname, _("Invalid value")))
        return errors

    def serialize(self, transport):
        """Transform received values if needed."""
        pass


class TransportBackendManager(object):
    """Transport backends manager."""

    def __init__(self):
        """Constructor."""
        self.backends = {}

    def get_backend(self, name, **kwargs):
        """Return backend instance."""
        backend_class = self.backends.get(name)
        if backend_class is None:
            return None
        return backend_class(**kwargs)

    def get_backend_list(self):
        """Return known backend list."""
        return sorted(
            (
                (key, key)
                for key, backend in self.backends.items()
            ),
            key=lambda i: i[1]
        )

    def get_backend_settings(self, name):
        """Return backend settings."""
        backend_class = self.backends.get(name)
        if backend_class is None:
            return None
        return backend_class.settings

    def get_all_backend_settings(self):
        """Return all backend settings."""
        return {
            name: backend_class.settings
            for name, backend_class in list(self.backends.items())
        }

    def register_backend(self, backend_class):
        """Register a new backend."""
        if not backend_class.name:
            raise RuntimeError("missing backend name")
        self.backends[backend_class.name] = backend_class


manager = TransportBackendManager()



# Generated by Django 1.10.5 on 2017-02-16 10:12
from django.db import migrations


def create_quota_limits(apps, schema_editor):
    """Create quota limits."""
    User = apps.get_model("core", "User")
    if not User.objects.exists():
        return
    ContentType = apps.get_model("contenttypes", "ContentType")
    UserObjectLimit = apps.get_model("limits", "UserObjectLimit")
    to_create = []
    ct = ContentType.objects.get(app_label="admin", model="domain")
    for u in User.objects.all():
        to_create.append(
            UserObjectLimit(user=u, name="quota", max_value=0, content_type=ct)
        )
    UserObjectLimit.objects.bulk_create(to_create)


class Migration(migrations.Migration):

    dependencies = [
        ('limits', '0005_auto_20160415_1654'),
    ]

    operations = [
        migrations.RunPython(create_quota_limits)
    ]



"""App related urls."""

from django.urls import path

from . import views

app_name = "dnstools"

urlpatterns = [
    path('dnsrecords/<int:pk>/', views.DNSRecordDetailView.as_view(),
         name="dns_record_detail"),
    path('dnsrecords/autoconfig/<int:pk>/',
         views.AutoConfigRecordsStatusView.as_view(),
         name="autoconfig_records_status"),
    path('configuration/<int:pk>/',
         views.DomainDNSConfigurationView.as_view(),
         name="domain_dns_configuration"),
]



"""Custom forms."""

from django import forms
from django.utils.translation import ugettext as _

from modoboa.lib.permissions import get_object_owner
from . import utils
from .lib import allocate_resources_from_user


class ResourcePoolForm(forms.Form):
    """Dynamic form to display user limits."""

    def __init__(self, *args, **kwargs):
        self.account = kwargs.pop("instance", None)
        super(ResourcePoolForm, self).__init__(*args, **kwargs)
        for name, tpl in utils.get_user_limit_templates():
            if "required_role" in tpl:
                condition = (
                    self.account is not None and
                    self.account.role != tpl["required_role"]
                )
                if condition:
                    continue
            self.fields["{}_limit".format(name)] = forms.IntegerField(
                label=tpl["label"], help_text=tpl["help"]
            )
        if self.account:
            self.load_from_user(self.account)

    def clean(self):
        """Ensure limit values are correct."""
        cleaned_data = super(ResourcePoolForm, self).clean()
        if self.errors:
            return cleaned_data
        for lname in list(self.fields.keys()):
            if cleaned_data[lname] < -1:
                self.add_error(lname, _("Invalid limit"))
        return cleaned_data

    def load_from_user(self, user):
        """Load limit values from given user."""
        for fieldname in list(self.fields.keys()):
            lname = fieldname.replace("_limit", "")
            self.fields[fieldname].initial = (
                user.userobjectlimit_set.get(name=lname).max_value)

    def save(self):
        owner = get_object_owner(self.account)
        for name, _definition in utils.get_user_limit_templates():
            fieldname = "{}_limit".format(name)
            if fieldname not in self.cleaned_data:
                continue
            limit = self.account.userobjectlimit_set.get(name=name)
            if not owner.is_superuser:
                allocate_resources_from_user(
                    limit, owner, self.cleaned_data[fieldname])
            limit.max_value = self.cleaned_data[fieldname]
            limit.save(update_fields=["max_value"])


class DomainLimitsForm(forms.Form):
    """Per-domain limits form."""

    def __init__(self, *args, **kwargs):
        """Define limits as fields."""
        self.domain = kwargs.pop("instance", None)
        super(DomainLimitsForm, self).__init__(*args, **kwargs)
        for name, tpl in utils.get_domain_limit_templates():
            self.fields["{}_limit".format(name)] = forms.IntegerField(
                label=tpl["label"], help_text=tpl["help"]
            )
        if not self.domain:
            return
        for fieldname in list(self.fields.keys()):
            lname = fieldname.replace("_limit", "")
            self.fields[fieldname].initial = (
                self.domain.domainobjectlimit_set.get(name=lname).max_value)

    def clean(self):
        """Ensure limit values are correct."""
        cleaned_data = super(DomainLimitsForm, self).clean()
        if self.errors:
            return cleaned_data
        for lname in list(self.fields.keys()):
            if cleaned_data[lname] < -1:
                self.add_error(lname, _("Invalid limit"))
        return cleaned_data

    def save(self, user, **kwargs):
        """Set limits."""
        for name, _definition in utils.get_domain_limit_templates():
            fieldname = "{}_limit".format(name)
            if fieldname not in self.cleaned_data:
                continue
            limit = self.domain.domainobjectlimit_set.get(name=name)
            limit.max_value = self.cleaned_data[fieldname]
            limit.save()



import os
from unittest import mock

import dns.resolver

from django.core.files.base import ContentFile
from django.core.management import call_command
from django.core.management.base import CommandError
from django.urls import reverse

from modoboa.core import factories as core_factories
from modoboa.core.models import User
from modoboa.lib.tests import ModoTestCase
from . import utils
from .. import factories
from ..models import Alias, Domain, DomainAlias


class ImportTestCase(ModoTestCase):

    @classmethod
    def setUpTestData(cls):  # NOQA:N802
        """Create test data."""
        super(ImportTestCase, cls).setUpTestData()
        cls.localconfig.parameters.set_value(
            "enable_admin_limits", False, app="limits")
        cls.localconfig.save()
        factories.populate_database()

    def test_domains_import(self):
        response = self.client.get(reverse("admin:domain_import"))
        self.assertEqual(response.status_code, 200)
        self.assertIn("Provide a CSV", response.content.decode())

        f = ContentFile(b"""domain; domain1.com; 1000; 100; True
domain; domain2.com; 1000; 200; False
domainalias; domalias1.com; domain1.com; True
""", name="domains.csv")
        self.client.post(
            reverse("admin:domain_import"), {
                "sourcefile": f
            }
        )
        admin = User.objects.get(username="admin")
        dom = Domain.objects.get(name="domain1.com")
        self.assertEqual(dom.quota, 1000)
        self.assertEqual(dom.default_mailbox_quota, 100)
        self.assertTrue(dom.enabled)
        self.assertTrue(admin.is_owner(dom))
        domalias = DomainAlias.objects.get(name="domalias1.com")
        self.assertEqual(domalias.target, dom)
        self.assertTrue(dom.enabled)
        self.assertTrue(admin.is_owner(domalias))
        dom = Domain.objects.get(name="domain2.com")
        self.assertEqual(dom.default_mailbox_quota, 200)
        self.assertFalse(dom.enabled)
        self.assertTrue(admin.is_owner(dom))

    def test_domain_import_bad_syntax(self):
        """Check errors handling."""
        url = reverse("admin:domain_import")
        f = ContentFile("domain; domain1.com; 100; True",
                        name="domains.csv")
        response = self.client.post(url, {"sourcefile": f})
        self.assertContains(response, "Invalid line")
        f = ContentFile("domain; domain1.com; XX; 100; True",
                        name="domains.csv")
        response = self.client.post(url, {"sourcefile": f})
        self.assertContains(response, "Invalid quota value")
        f = ContentFile("domain; domain1.com; 100; XX; True",
                        name="domains.csv")
        response = self.client.post(url, {"sourcefile": f})
        self.assertContains(response, "Invalid default mailbox quota")
        f = ContentFile("domain; domain1.com; 10; 100; True",
                        name="domains.csv")
        response = self.client.post(url, {"sourcefile": f})
        self.assertContains(
            response,
            "Default mailbox quota cannot be greater than domain quota")

    @mock.patch.object(dns.resolver.Resolver, "resolve")
    @mock.patch("socket.getaddrinfo")
    def test_domain_import_with_mx_check(self, mock_getaddrinfo, mock_query):
        """Check domain import when MX check is enabled."""
        reseller = core_factories.UserFactory(
            username="reseller", groups=("Resellers", ))
        self.client.force_login(reseller)
        self.set_global_parameter("valid_mxs", "192.0.2.1 2001:db8::1")
        self.set_global_parameter("domains_must_have_authorized_mx", True)

        mock_query.side_effect = utils.mock_dns_query_result
        mock_getaddrinfo.side_effect = utils.mock_ip_query_result
        f = ContentFile(
            b"domain; test3.com; 100; 1; True", name="domains.csv")
        resp = self.client.post(
            reverse("admin:domain_import"), {
                "sourcefile": f
            }
        )
        self.assertContains(resp, "No authorized MX record found for domain")

        mock_getaddrinfo.side_effect = utils.mock_ip_query_result
        f = ContentFile(
            b"domain; domain1.com; 100; 1; True", name="domains.csv")
        resp = self.client.post(
            reverse("admin:domain_import"), {
                "sourcefile": f
            }
        )
        self.assertTrue(
            Domain.objects.filter(name="domain1.com").exists())

    def test_import_domains_with_conflict(self):
        f = ContentFile(b"""domain;test.alias;100;10;True
domainalias;test.alias;test.com;True
""", name="domains.csv")
        resp = self.client.post(
            reverse("admin:domain_import"), {
                "sourcefile": f
            }
        )
        self.assertIn(
            "Object already exists: domainalias", resp.content.decode())

    def test_identities_import(self):
        response = self.client.get(reverse("admin:identity_import"))
        self.assertEqual(response.status_code, 200)
        self.assertIn("Provide a CSV", response.content.decode())

        f = ContentFile("""
account; user1@test.com; toto; User; One; True; SimpleUsers; user1@test.com; 0
account; Truc@test.com; toto; René; Truc; True; DomainAdmins; truc@test.com; 5; test.com
alias; alias1@test.com; True; user1@test.com
forward; alias2@test.com; True; user1+ext@test.com
forward; fwd1@test.com; True; user@extdomain.com
dlist; dlist@test.com; True; user1@test.com; user@extdomain.com
""", name="identities.csv")  # NOQA:E501
        self.client.post(
            reverse("admin:identity_import"),
            {"sourcefile": f, "crypt_password": True}
        )
        admin = User.objects.get(username="admin")
        u1 = User.objects.get(username="user1@test.com")
        mb1 = u1.mailbox
        self.assertTrue(admin.is_owner(u1))
        self.assertEqual(u1.email, "user1@test.com")
        self.assertEqual(u1.first_name, "User")
        self.assertEqual(u1.last_name, "One")
        self.assertTrue(u1.is_active)
        self.assertEqual(u1.role, "SimpleUsers")
        self.assertTrue(mb1.use_domain_quota)
        self.assertEqual(mb1.quota, 0)
        self.assertTrue(admin.is_owner(mb1))
        self.assertEqual(mb1.full_address, "user1@test.com")
        self.assertTrue(
            self.client.login(username="user1@test.com", password="toto")
        )

        da = User.objects.get(username="truc@test.com")
        damb = da.mailbox
        self.assertEqual(da.first_name, u"René")
        self.assertEqual(da.role, "DomainAdmins")
        self.assertEqual(damb.quota, 5)
        self.assertFalse(damb.use_domain_quota)
        self.assertEqual(damb.full_address, "truc@test.com")
        dom = Domain.objects.get(name="test.com")
        self.assertIn(da, dom.admins)
        u = User.objects.get(username="user@test.com")
        self.assertTrue(da.can_access(u))

        al = Alias.objects.get(address="alias1@test.com")
        self.assertTrue(
            al.aliasrecipient_set
            .filter(r_mailbox=u1.mailbox).exists()
        )
        self.assertTrue(admin.is_owner(al))

        fwd = Alias.objects.get(address="fwd1@test.com")
        self.assertTrue(
            fwd.aliasrecipient_set
            .filter(
                address="user@extdomain.com", r_mailbox__isnull=True,
                r_alias__isnull=True)
            .exists()
        )
        self.assertTrue(admin.is_owner(fwd))

        dlist = Alias.objects.get(address="dlist@test.com")
        self.assertTrue(
            dlist.aliasrecipient_set
            .filter(r_mailbox=u1.mailbox).exists()
        )
        self.assertTrue(
            dlist.aliasrecipient_set.filter(address="user@extdomain.com")
            .exists()
        )
        self.assertTrue(admin.is_owner(dlist))

    def test_import_for_nonlocal_domain(self):
        """Try to import an account for nonlocal domain."""
        f = ContentFile(b"""
account; user1@nonlocal.com; toto; User; One; True; SimpleUsers; user1@nonlocal.com; 0
""", name="identities.csv")  # NOQA:E501
        self.client.post(
            reverse("admin:identity_import"),
            {"sourcefile": f, "crypt_password": True}
        )
        self.assertFalse(
            User.objects.filter(username="user1@nonlocal.com").exists())

    def test_import_invalid_quota(self):
        f = ContentFile(b"""
account; user1@test.com; toto; User; One; True; SimpleUsers; user1@test.com; ; test.com
""", name="identities.csv")  # NOQA:E501
        resp = self.client.post(
            reverse("admin:identity_import"),
            {"sourcefile": f, "crypt_password": True}
        )
        self.assertIn("wrong quota value", resp.content.decode())

    def test_import_domain_by_domainadmin(self):
        """Check if a domain admin is not allowed to import a domain."""
        self.client.logout()
        self.client.login(username="admin@test.com", password="toto")
        f = ContentFile(b"""
domain; domain2.com; 1000; 200; False
""", name="identities.csv")
        resp = self.client.post(
            reverse("admin:identity_import"),
            {"sourcefile": f, "crypt_password": True}
        )
        self.assertIn(
            "You are not allowed to import domains",
            resp.content.decode())
        f = ContentFile(b"""
domainalias; domalias1.com; test.com; True
""", name="identities.csv")
        resp = self.client.post(
            reverse("admin:identity_import"),
            {"sourcefile": f, "crypt_password": True}
        )
        self.assertIn(
            "You are not allowed to import domain aliases",
            resp.content.decode())

    def test_import_quota_too_big(self):
        self.client.logout()
        self.client.login(username="admin@test.com", password="toto")
        f = ContentFile(b"""
account; user1@test.com; toto; User; One; True; SimpleUsers; user1@test.com; 40
""", name="identities.csv")
        resp = self.client.post(
            reverse("admin:identity_import"),
            {"sourcefile": f, "crypt_password": True}
        )
        self.assertIn("Domain quota exceeded", resp.content.decode())

    def test_import_missing_quota(self):
        f = ContentFile(b"""
account; user1@test.com; toto; User; One; True; SimpleUsers; user1@test.com
""", name="identities.csv")
        self.client.post(
            reverse("admin:identity_import"),
            {"sourcefile": f, "crypt_password": True}
        )
        account = User.objects.get(username="user1@test.com")
        self.assertEqual(
            account.mailbox.quota,
            account.mailbox.domain.default_mailbox_quota
        )

    def test_import_duplicate(self):
        f = ContentFile("""
account; admin@test.com; toto; Admin; ; True; DomainAdmins; admin@test.com; 0; test.com
account; truc@test.com; toto; René; Truc; True; DomainAdmins; truc@test.com; 0; test.com
""", name="identities.csv")  # NOQA:E501
        self.client.post(
            reverse("admin:identity_import"),
            {"sourcefile": f, "crypt_password": True,
             "continue_if_exists": True}
        )
        admin = User.objects.get(username="admin")
        u1 = User.objects.get(username="truc@test.com")
        self.assertTrue(admin.is_owner(u1))

    def test_import_superadmin(self):
        """Check if a domain admin can import a superadmin

        Expected result: no
        """
        self.client.logout()
        self.assertTrue(
            self.client.login(username="admin@test.com", password="toto")
        )
        f = ContentFile(b"""
account; sa@test.com; toto; Super; Admin; True; SuperAdmins; superadmin@test.com; 50
""", name="identities.csv")  # NOQA:E501
        self.client.post(
            reverse("admin:identity_import"),
            {"sourcefile": f, "crypt_password": True,
             "continue_if_exists": True}
        )
        with self.assertRaises(User.DoesNotExist):
            User.objects.get(username="sa@test.com")

    def test_import_alias_with_empty_values(self):
        f = ContentFile(b"""
alias;user.alias@test.com;True;user@test.com;;;;;;;;;;;;;;;;
""", name="identities.csv")
        self.client.post(
            reverse("admin:identity_import"),
            {"sourcefile": f, "crypt_password": True,
             "continue_if_exists": True}
        )
        alias = Alias.objects.get(address="user.alias@test.com")
        self.assertEqual(alias.type, "alias")

    def test_import_account_alias_conflict(self):
        """Specific test for #1144."""
        f = ContentFile(b"""
alias;user@test.com;True;admin@test.com
""", name="identities.csv")
        self.client.post(
            reverse("admin:identity_import"),
            {"sourcefile": f, "crypt_password": True}
        )
        self.assertTrue(
            Alias.objects.filter(
                address="user@test.com", internal=False).exists())

    def test_domains_import_utf8(self):
        response = self.client.get(reverse("admin:domain_import"))
        self.assertEqual(response.status_code, 200)
        self.assertIn("Provide a CSV", response.content.decode())

        f = ContentFile("""domain; dómªin1.com; 1000; 100; True
dómain; dómªin2.com; 1000; 100; True
""".encode("utf8"), name="dómains.csv")
        self.client.post(
            reverse("admin:domain_import"), {
                "sourcefile": f
            }
        )
        admin = User.objects.get(username="admin")
        dom = Domain.objects.get(name="dómªin1.com")
        self.assertEqual(dom.quota, 1000)
        self.assertEqual(dom.default_mailbox_quota, 100)
        self.assertTrue(dom.enabled)
        self.assertTrue(admin.is_owner(dom))

    def test_import_command_missing_file(self):
        with mock.patch("os.path.isfile") as mock_isfile:
            mock_isfile.return_value = False
            with self.assertRaises(CommandError) as cm:
                call_command("modo", "import", "test.csv")
                ex_message = cm.exception.messages
                self.assertEqual(ex_message, "File not found")

    def test_import_command(self):
        test_file = os.path.join(
            os.path.dirname(__file__),
            "test_data/import_domains.csv"
        )
        call_command("modo", "import", test_file)

        admin = User.objects.get(username="admin")

        dom = Domain.objects.get(name="domain1.com")
        self.assertEqual(dom.quota, 1000)
        self.assertEqual(dom.default_mailbox_quota, 100)
        self.assertTrue(dom.enabled)
        self.assertTrue(admin.is_owner(dom))

        dom = Domain.objects.get(name="dómain2.com")
        self.assertEqual(dom.quota, 2000)
        self.assertEqual(dom.default_mailbox_quota, 200)
        self.assertTrue(dom.enabled)
        self.assertTrue(admin.is_owner(dom))

    def test_import_command_non_utf8(self):
        test_file = os.path.join(
            os.path.dirname(__file__),
            "test_data/import_domains_iso8859.csv"
        )
        call_command("modo", "import", test_file)
        dom = Domain.objects.get(name="dómain2.com")
        self.assertEqual(dom.quota, 2000)

    def test_import_command_duplicates(self):
        test_file = os.path.join(
            os.path.dirname(__file__),
            "test_data/import_domains_duplicates.csv"
        )
        with self.assertRaises(CommandError) as cm:
            call_command("modo", "import", test_file)
            ex_message = cm.exception.messages
            self.assertTrue(ex_message.startswith("Object already exists: "))

    def test_import_command_duplicates_continue(self):
        test_file = os.path.join(
            os.path.dirname(__file__),
            "test_data/import_domains_duplicates.csv"
        )
        call_command("modo", "import", "--continue-if-exists", test_file)

        admin = User.objects.get(username="admin")

        dom = Domain.objects.get(name="domain1.com")
        self.assertEqual(dom.quota, 1000)
        self.assertEqual(dom.default_mailbox_quota, 100)
        self.assertTrue(dom.enabled)
        self.assertTrue(admin.is_owner(dom))

        dom = Domain.objects.get(name="dómain2.com")
        self.assertEqual(dom.quota, 2000)
        self.assertEqual(dom.default_mailbox_quota, 200)
        self.assertTrue(dom.enabled)
        self.assertTrue(admin.is_owner(dom))
